"""
This file deals with all of the email functions.
"""

import logging

from flask import render_template, current_app
from flask_mail import Message

from ..app import db_session
from ..proxies import current_repo
from ..models import Email, Subscription, User, Post
from ..utils.render import render_post

logger = logging.getLogger(__name__)

# TODO move me to config
TAGS_EXCLUDED_FROM_SUBSCRIPTION_EMAILS = ['other/private', 'private']


def usernames_to_emails(usernames):
    username_to_email = current_repo.config.username_to_email
    return [username_to_email(username) for username in usernames]


def subscription_email_recipients(post, tag):
    """Check who should recieve a subscription email about this post
    """
    db_session.expire_on_commit = False

    subscriptions = (db_session.query(Subscription)
                               .filter(Subscription.object_type == 'tag')
                               .filter(Subscription.object_id == tag.id)
                               .all())

    subscriber_ids = list(set([s.user_id for s in subscriptions]))  # uniqify
    emails_sent = (db_session.query(Email)
                             .filter(Email.object_id == post.id)
                             .all())

    users_rcvd_emails = [e.user_id for e in emails_sent]
    new_email_recipient_ids = [
        user_id for user_id in subscriber_ids if user_id not in users_rcvd_emails]

    if not new_email_recipient_ids:
        return []
    new_email_recipients = (db_session.query(User)
                                      .filter(User.id.in_(new_email_recipient_ids))
                                      .all())
    return new_email_recipients


def send_subscription_emails(post):
    """ Send an email to all those subscribed to a tag if a new post
        with that tag was published to the repo
    """
    if 'mail' not in current_app.config:
        logger.warning(
            'Mail subsystem is not configured. Silently dropping email.')
        return

    # if this post is tagged as private - send no emails
    post_full_tags = [tag.name for tag in post.tags]
    for full_tag in post_full_tags:
        if full_tag in TAGS_EXCLUDED_FROM_SUBSCRIPTION_EMAILS:
            return
    for tag in post.tags:
        send_subscription_email(post, tag)


def send_subscription_email(post, tag):
    """
    Send an email to all those subscribed to a tag if a new post
    with that tag was published to the repo

    :param post: The new post that the email is sent about
    :type post: The Post object
    :param tag: The tag that people are subscribed to
    :type tag: The Tag object
    :return: No return
    :rtype: None
    """
    # format recipients for email and mark as sent
    recipient_users = subscription_email_recipients(post, tag)

    if not recipient_users:
        return

    recipients_bcc = [usernames_to_emails([user.username])[0] for user in recipient_users]

    if not recipients_bcc:
        return

    default_recipients = ['knowledge_consumer@notreal.com']
    subject = "New knowledge post with tag [{}]!".format(tag.name)
    msg = Message(subject=subject,
                  recipients=default_recipients,
                  bcc=recipients_bcc)

    # Iterate over fetched image files and attach them to email
    post_text = render_post(post)

    msg.body = render_template("email_templates/subscription_email.txt",
                               full_tag=tag.name,
                               page_id=post.path,
                               post_text=post_text,
                               knowledge_app_base_url=current_app.config['SERVER_NAME'])

    msg.html = post_text

    for user in recipient_users:
        # mark email as sent just before you send the email
        email_sent = Email(user_id=user.id,
                           trigger_id=tag.id,
                           trigger_type='subscription',
                           object_id=post.id,
                           object_type="post",
                           subject=subject,
                           text=post_text)
        db_session.add(email_sent)
        db_session.commit()

    current_app.config['mail'].send(msg)


def send_comment_email(post_id, comment_text, commenter='Someone'):
    if 'mail' not in current_app.config:
        logger.warning(
            'Mail subsystem is not configured. Silently dropping email.')
        return

    gitpost = (db_session.query(Post)
               .filter(Post.id == post_id)
               .first())
    post_title = gitpost.title
    authors = [author.username for author in gitpost.authors]
    post_author_emails = usernames_to_emails(authors)

    msg = Message("Someone commented on your post!", post_author_emails)
    msg.body = render_template("email_templates/comment_email.txt",
                               commenter=commenter,
                               comment_text=comment_text,
                               post_title=post_title,
                               post_id=post_id)
    current_app.config['mail'].send(msg)


def send_internal_error_email(subject_line, **kwargs):
    if 'mail' not in current_app.config:
        logger.warning(
            'Mail subsystem is not configured. Silently dropping email.')
        return
    recipients = usernames_to_emails(current_repo.config.editors)
    msg = Message(subject_line, recipients)
    msg.body = render_template(
        "email_templates/internal_error_email.txt", **kwargs)
    current_app.config['mail'].send(msg)


def send_reviewer_request_email(post_id, reviewer):
    if 'mail' not in current_app.config:
        logger.warning('Mail subsystem is not configured. Silently dropping email.')
        return
    subject = "Someone requested a web post review from you!"
    msg = Message(subject, [reviewer])
    msg.body = render_template("email_templates/reviewer_request_email.txt",
                               post_id=post_id)
    current_app.config['mail'].send(msg)


def send_review_email(post_id, comment_text, commenter='Someone'):
    if 'mail' not in current_app.config:
        logger.warning('Mail subsystem is not configured. Silently dropping email.')
        return
    webpost = (db_session.query(Post)
               .filter(Post.id == post_id)
               .first())
    post_title = webpost.title
    authors = [author.username for author in webpost.authors]
    post_author_emails = usernames_to_emails(authors)
    subject = "Someone reviewed your web post!"
    msg = Message(subject, post_author_emails)
    msg.body = render_template("email_templates/review_email.txt",
                               commenter=commenter,
                               comment_text=comment_text,
                               post_title=post_title,
                               post_id=post_id)
    current_app.config['mail'].send(msg)



# Configuration for this knowledge data repository.


# A function called to see whether a specified path is permitted in the repository
# Only enforced when creating/modifying posts. It should return the path as a standard
# unix path (virtual folders separated by '/' and the final node should end in '.kp').
# It should raise an exception if the provided path is is not permitted in the knowledge
# repository.
def path_parse(path):
    return path


# A dictionary of aliases which point to knowledge posts. This allows you to alias posts
# which may be useful to forward old links to new locations, or to give certain posts
# precedence. It also allows the post path to violate the normal naming rules of posts.
# The only condition is that they cannot mask an existing post url, in which case the
# alias is silently ignored.
aliases = {}


# Postprocessors to apply when importing KnowledgePost objects into the repository.
# Note that KnowledgePost objects by default run 'extract_images' and 'format_checks'.
# Order is important.
postprocessors = []


# Usernames of users to keep informed of changes to the knowledge data repo
editors = []


# Function to check whether provided username is a valid username, and if not, mutate it
# such that it is. Should raise an exception if that is not possible, and otherwise
# return the parsed username.
def username_parse(username):
    return username


# Function to convert a username to a person's proper name
def username_to_name(username):
    return username


# Function to convert a username to a person's email
def username_to_email(username):
    return '{}@example.com'.format(username)


# Function to generate the web uri for a knowledge post at
# path `path`. If `None`, should return the web uri for
# the entire repository. Return `None` if a web uri does
# not exist.
def web_uri(path=None):
    return None


# WARNING: ADVANCED!
# Function that is called on a Flask web app instance before it is launched
# This is useful if you need to add a route for your local deployment, or other
# equally invasive action. Not recommended unless you know exactly what you are doing.
# It must return a KnowledgeFlask app instance.
def prepare_app(app):
    return app



import unittest

from knowledge_repo import KnowledgeRepository


class RouteTest(unittest.TestCase):

    def setUp(self):
        self.repo = KnowledgeRepository.for_uri('tests/test_repo', auto_create=True)
        self.app = self.repo.get_app(config='tests/config_server.py').test_client()
        self.headers = {}

    def test01_test_index(self):
        """
        Test routing / to /feed
        """
        rv = self.app.get('/', headers=self.headers)
        # index reroutes to /feed
        assert (rv.status == "302 FOUND")

    def test02_test_health(self):
        """
        Test /health
        """
        rv = self.app.get("/health", headers=self.headers)
        assert (rv.status == "200 OK")
        assert (rv.data.decode('utf-8') == "OK")

    def test03_test_ping(self):
        """
        Test /ping
        """
        rv = self.app.get("/ping", headers=self.headers)
        assert (rv.status == "200 OK")
        assert (rv.data.decode('utf-8') == "OK")

    def test04_test_stats(self):
        """
        Test /stats
        """
        rv = self.app.get('/stats')
        assert rv.status == '200 OK'

if __name__ == '__main__':
    unittest.main()



from ._version import *
from .utils.dependencies import check_dependencies
check_dependencies(__dependencies__)

from .repository import KnowledgeRepository  # nopep8
from .post import KnowledgePost  # nopep8

from . import converters  # nopep8
from . import postprocessors  # nopep8
from . import repositories  # nopep8

import logging  # nopep8
logging.basicConfig(level=logging.INFO)



from __future__ import absolute_import
from __future__ import unicode_literals
import markdown
from markdown import Extension
from markdown.preprocessors import Preprocessor
import re
import base64
import mimetypes

from ..converter import KnowledgePostConverter
from ..mapping import SubstitutionMapper

MARKDOWN_EXTENSTIONS = ['markdown.extensions.extra',
                        'markdown.extensions.abbr',
                        'markdown.extensions.attr_list',
                        'markdown.extensions.def_list',
                        'markdown.extensions.fenced_code',
                        'markdown.extensions.footnotes',
                        'markdown.extensions.tables',
                        'markdown.extensions.smart_strong',
                        'markdown.extensions.admonition',
                        'markdown.extensions.codehilite',
                        'markdown.extensions.headerid',
                        'markdown.extensions.meta',
                        'markdown.extensions.sane_lists',
                        'markdown.extensions.smarty',
                        'markdown.extensions.toc(baselevel=3)',
                        'markdown.extensions.wikilinks',
                        'knowledge_repo.converters.html:KnowledgeMetaExtension',
                        'knowledge_repo.converters.html:MathJaxExtension']


class KnowledgeMetaPreprocessor(Preprocessor):
    """ Get Meta-Data. """

    def run(self, lines):
        """ Parse Meta-Data and store in Markdown.Meta. """
        cnt = 0
        for i, line in enumerate(lines):
            if line.strip() == '---':
                cnt = cnt + 1
            if cnt == 2:
                break
        return lines[i + 1:]


class KnowledgeMetaExtension(Extension):
    """ Meta-Data extension for Python-Markdown. """

    def extendMarkdown(self, md, md_globals):
        """ Add MetaPreprocessor to Markdown instance. """
        md.preprocessors.add("knowledge_meta",
                             KnowledgeMetaPreprocessor(md),
                             ">normalize_whitespace")


class MathJaxPattern(markdown.inlinepatterns.Pattern):

    def __init__(self):
        markdown.inlinepatterns.Pattern.__init__(self, r'(?<!\\)(\$\$?)(.+?)\2')

    def handleMatch(self, m):
        node = markdown.util.etree.Element('mathjax')
        node.text = markdown.util.AtomicString(m.group(2) + m.group(3) + m.group(2))
        return node


class MathJaxExtension(markdown.Extension):
    def extendMarkdown(self, md, md_globals):
        # Needs to come before escape matching because \ is pretty important in LaTeX
        md.inlinePatterns.add('mathjax', MathJaxPattern(), '<escape')


class HTMLConverter(KnowledgePostConverter):
    '''
    Use this as a template for new KnowledgePostConverters.
    '''
    _registry_keys = ['html']

    def init(self):
        self.kp_images = self.kp.read_images()

    def to_string(self, skip_headers=False, images_base64_encode=True, urlmappers=[]):
        """
        Renders the markdown as html
        """
        # Copy urlmappers locally so we can modify it without affecting global
        # state
        urlmappers = urlmappers[:]
        if images_base64_encode:
            urlmappers.insert(0, self.base64_encode_image_mapper)

        html = ''
        if not skip_headers:
            html += self.render_headers()
        html += markdown.Markdown(extensions=MARKDOWN_EXTENSTIONS).convert(self.kp.read())

        return self.apply_url_remapping(html, urlmappers)

    def apply_url_remapping(self, html, urlmappers):
        patterns = {
            'img': '<img.*?src=[\'"](?P<url>.*?)[\'"].*?>',
            'a': '<a.*?href=[\'"](?P<url>.*?)[\'"].*?>'
        }

        def urlmapper_proxy(name, match):
            for urlmapper in urlmappers:
                new_url = urlmapper(name, match.group('url'))
                if new_url is not None:
                    break
            if new_url is not None:
                return re.sub('(src|href)=[\'"](?:.*?)[\'"]', '\\1="{}"'.format(new_url), match.group(0))
            return None

        return SubstitutionMapper(patterns=patterns, mappers=[urlmapper_proxy]).apply(html)

    # Utility methods
    def render_headers(self):
        headers = self.kp.headers

        headers['authors_string'] = ', '.join(headers.get('authors'))
        headers['tldr'] = markdown.Markdown(extensions=MARKDOWN_EXTENSTIONS[
                                            :-1]).convert(headers['tldr'])
        headers['date_created'] = headers['created_at'].isoformat()
        headers['date_updated'] = headers['updated_at'].isoformat()

        header = """
<h1>{title}</h1>
<p id='metadata'>
<strong>Author</strong>: {authors_string} <br>
<strong>Date Created</strong>: {date_created}<br>
<strong>Date Updated</strong>: {date_updated}<br>
<strong>Tags</strong><text>: </text><br>
<strong>TLDR</strong>: {tldr}<br>
</p>
""".format(**headers)

        return header

    def base64_encode_image_mapper(self, tag, url):
        if tag == 'img':
            if url in self.kp_images:
                image_data = base64.b64encode(self.kp_images[url])
                image_mimetype = mimetypes.guess_type(url)[0]
                if image_mimetype is not None:
                    return 'data:{};base64, '.format(image_mimetype) + image_data.decode('utf-8')
        return None



import sys
import imp
import time


def get_module_for_source(source, module_name, register_globally=True):
    module_name = module_name + '_' + str(time.time()).replace('.', '')
    module = imp.new_module(module_name)
    exec(source, module.__dict__)
    if register_globally:
        sys.modules[module_name] = module
    return module



# This file contains code that is largely copied from
# https://bitbucket.org/zzzeek/sqlalchemy/wiki/UsageRecipes/UniqueObject

from ..proxies import db_session


def unique_cache(session, cls, hashfunc, queryfunc, constructor, arg, kw):
    cache = getattr(session, '_unique_cache', None)
    if cache is None:
        session._unique_cache = cache = {}

    key = (cls, hashfunc(*arg, **kw))
    if key in cache:
        return cache[key]
    else:
        with session.no_autoflush:
            q = session.query(cls)
            q = queryfunc(q, *arg, **kw)
            obj = q.first()
            if not obj:
                obj = constructor(*arg, **kw)
                session.add(obj)
        cache[key] = obj
        return obj


def unique_constructor(hashfunc, queryfunc):
    def decorate(cls):
        def _null_init(self, *arg, **kw):
            pass

        def __new__(cls, bases, *arg, **kw):
            # no-op __new__(), called
            # by the loading procedure
            if not arg and not kw:
                return object.__new__(cls)

            session = db_session()

            def constructor(*arg, **kw):
                obj = object.__new__(cls)
                obj._init(*arg, **kw)
                return obj

            return unique_cache(
                session,
                cls,
                hashfunc,
                queryfunc,
                constructor,
                arg, kw
            )

        # note: cls must be already mapped for this part to work
        cls._init = cls.__init__
        cls.__init__ = _null_init
        cls.__new__ = classmethod(__new__)
        return cls

    return decorate



"""Functions that interact with posts.

Functions include:
    - get_posts
    - get_all_post_stats
"""
from sqlalchemy import func, distinct, or_

from ..app import db_session
from ..models import (Comment, PageView, Post,
                      Tag, Vote)


def get_query_param_set(params):
    """
    Strip, lowercase, and remove empty params to be used in a query
    """
    param_set = params.strip().lower().split(" ")
    param_set = [p for p in param_set if len(p) > 0]
    return param_set


def get_posts(feed_params):
    """
    Return a list of post objects (either WebEditorPosts or GitPosts)
    by building a query based on the feed_params
    :param feed_params: Parameters in the url request
    :type feed_params: object
    :return: Posts matching feed param specification
    :rtype: Tuple
    """

    # make sure post is published
    query = (db_session.query(Post).filter(Post.is_published))

    # if a private tag exists, make sure that post
    private_tag = (db_session.query(Tag)
                   .filter(Tag.name == 'other/private')
                   .first())
    if private_tag is not None:
        query = query.filter(~Post.tags.contains(private_tag))

    # filter out based on feed param filters
    filters = feed_params['filters']
    if filters and str(filters):
        filter_set = get_query_param_set(filters)
        for elem in filter_set:
            query = query.filter(or_(Post.title.like('%' + elem + '%'),
                                     Post.title.like('%' + elem),
                                     Post.title.like(elem + '%'),
                                     Post.tldr.like('%' + elem + '%'),
                                     Post.tldr.like('%' + elem),
                                     Post.tldr.like(elem + '%')))

    # sort - TODO clean up
    sort_by = feed_params['sort_by']

    # sort by post property
    post_properties = {
        "updated_at": Post.updated_at,
        "created_at": Post.created_at,
        "title": Post.title,
    }
    join_order_col = {
        "uniqueviews": func.count(distinct(PageView.user_id)),
        "allviews": func.count(PageView.object_id),
        "upvotes": func.count(Vote.object_id),
        "comments": func.count(Comment.post_id)
    }

    order_col = None
    if sort_by in post_properties:
        order_col = post_properties[sort_by]
    elif sort_by in join_order_col:  # sort by joined property
        order_col = join_order_col[sort_by]

        joins = {
            "uniqueviews": (PageView, PageView.object_id),
            "allviews": (PageView, PageView.object_id),
            "upvotes": (Vote, Vote.object_id),
            "comments": (Comment, Comment.post_id)
        }

        (join_table, join_on) = joins[sort_by]

        query = (db_session.query(Post, order_col)
                 .outerjoin(join_table, Post.path == join_on))
        query = query.group_by(Post.path)

    # sort order
    if order_col is not None:
        if feed_params['sort_desc']:
            query = query.order_by(order_col.desc())
        else:
            query = query.order_by(order_col.asc())

    query = (query.order_by(Post.id.desc()))
    posts = query.all()

    # Check if a grouped by result, and if so, unnest Post object
    if posts and not isinstance(posts[0], Post):
        posts = [post[0] for post in posts]

    # Filter by authors TODO: this should be done by a hybrid_property
    author_names = feed_params['authors']
    if author_names:
        author_names = set(author_names.split(','))
        posts = [p for p in posts if author_names.intersection(
            set([auth.username for auth in p.authors]))]

    # get the right indexes
    posts = posts[feed_params['start']:feed_params[
        'start'] + feed_params['results']]

    # Post.authors is lazy loaded, so we need to make sure it has been loaded before being
    # passed beyond the scope of this database db_session.
    for post in posts:
        post.authors

    post_stats = {post.path: {'all_views': post.view_count,
                              'distinct_views': post.view_user_count,
                              'total_likes': post.vote_count,
                              'total_comments': post.comment_count} for post in posts}

    db_session.expunge_all()

    return posts, post_stats



import unittest

from bs4 import BeautifulSoup
from sqlalchemy import and_

from knowledge_repo import KnowledgeRepository
from knowledge_repo.app.proxies import db_session
from knowledge_repo.app.models import Post


class PostTest(unittest.TestCase):

    def setUp(self):
        self.repo = KnowledgeRepository.for_uri('tests/test_repo', auto_create=True)
        self.app = self.repo.get_app(config='tests/config_server.py')
        self.client = self.app.test_client()

        with self.app.app_context():
            # for this test, pick one post
            post = (db_session.query(Post)
                              .filter(Post.is_published)
                              .first())
            self.post_id = post.id
            self.post_rendered_url = '/render?markdown={}'.format(post.path)

    def test01_check_render_route(self):
        """
        Check post path will render in /render?markdown=
        """
        rv = self.client.get(self.post_rendered_url)
        assert rv.status == "200 OK"

    def test02_check_post_stats(self):
        """
        Make sure pageview, distinct viewers correct for post
        """
        rv = self.client.get(self.post_rendered_url)

        data = rv.data.decode('utf-8')
        soup = BeautifulSoup(data, 'html.parser')

        # we'll check the post_stats of the post by computing them
        # and ensuring that the upper-right-hand text is correct
        icon = soup.findAll("div", {"id": "pageview_stats"})
        pageviews_str = icon[0].text.strip()

        with self.app.app_context():
            post = (db_session.query(Post)
                              .filter(Post.id == self.post_id)
                              .first())
            distinct_viewers = post.view_user_count
            pageviews = post.view_count

        db_pageviews_str = "Viewed {pageviews} times by {distinct_viewers} different users".format(**locals())

        assert db_pageviews_str == pageviews_str, "'{}' is not '{}'!".format(db_pageviews_str, pageviews_str)

    def test03_check_post_metadata(self):
        """
        Check post metadata
        """
        rv = self.client.get(self.post_rendered_url)

        data = rv.data.decode('utf-8')
        soup = BeautifulSoup(data, 'html.parser')
        metadata = soup.findAll("p", {"id": "metadata"})

        assert metadata

        metadata = metadata[0]
        children = metadata.contents

        # strip out the \n
        children = [c for c in children if c != ' ' and c != '\n']

        for (i, child) in enumerate(children):
            try:
                metadata_elem = child.findAll("strong")
                assert children[i + 1].contents
            except:
                pass

if __name__ == '__main__':
    unittest.main()



import os
import sys
import traceback
from future.utils import raise_with_traceback
from flask import current_app, request, g
from flask_sqlalchemy import SQLAlchemy
import functools
from collections import defaultdict

from werkzeug.local import LocalProxy
from sqlalchemy import func, distinct, and_, select
import logging

from knowledge_repo._version import __version__
from knowledge_repo.repository import KnowledgeRepository
from .proxies import current_repo, db_session
from .utils.models import unique_constructor
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.ext.associationproxy import association_proxy

logger = logging.getLogger(__name__)

db = SQLAlchemy()
db_session = LocalProxy(lambda: current_app.db.session)


class PostAuthorAssoc(db.Model):
    __tablename__ = 'assoc_post_author'

    post_id = db.Column(db.Integer, db.ForeignKey("posts.id"), nullable=False, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False, primary_key=True)
    order = db.Column(db.Integer)

    post = db.relationship('Post', lazy='joined')
    author = db.relationship('User', lazy='joined')


assoc_post_tag = db.Table(
    'assoc_post_tag',
    db.Model.metadata,
    db.Column('post_id', db.Integer, db.ForeignKey('posts.id')),
    db.Column('tag_id', db.Integer, db.ForeignKey('tags.id'))
)


class Comment(db.Model):
    __tablename__ = 'comments'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer)
    post_id = db.Column(db.Integer)
    text = db.Column(db.Text)
    type = db.Column(db.String(100), default='post')
    created_at = db.Column(db.DateTime, default=func.now())
    updated_at = db.Column(db.DateTime, default=func.now(), onupdate=func.now())


class PageView(db.Model):
    __tablename__ = 'pageviews'

    id = db.Column(db.Integer, primary_key=True)
    page = db.Column(db.String(512))
    endpoint = db.Column(db.String(255))
    user_id = db.Column(db.Integer)
    object_id = db.Column(db.Integer)
    object_type = db.Column(db.String(100))
    object_action = db.Column(db.String(100))
    ip_address = db.Column(db.String(64))
    created_at = db.Column(db.DateTime, default=func.now())
    error_message = db.Column(db.Text())
    version = db.Column(db.String(100))

    class logged(object):

        def __init__(self, route, object_extractor=None):
            self._route = route
            self._object_extractor = object_extractor

        def __getattr__(self, attr):
            return getattr(self._route, attr)

        def __call__(self, *args, **kwargs):
            if not current_app.config.get('REPOSITORY_INDEXING_ENABLED', True):
                return self._route(*args, **kwargs)

            log = PageView(
                page=request.full_path,
                endpoint=request.endpoint,
                user_id=g.user.id,
                ip_address=request.remote_addr,
                version=__version__
            )
            log.object_id, log.object_type, log.object_action, reextract_after_request = self.extract_objects(*args, **kwargs)
            db_session.add(log)

            try:
                return self._route(*args, **kwargs)
            except Exception as e:
                tb = traceback.extract_tb(sys.exc_info()[2])
                log.error_message = type(e).__name__ + ': ' + str(e) + '\nTraceback (most recent call last):\n' + '\n'.join(traceback.format_list(tb[1:]))
                raise_with_traceback(e)
            finally:
                # Extract object id and type after response generated (if requested) to ensure
                # most recent data is collected
                if reextract_after_request:
                    log.object_id, log.object_type, log.object_action, _ = self.extract_objects(*args, **kwargs)

                db_session.rollback()
                db_session.add(log)
                db_session.commit()

        def object_extractor(self, extractor):
            self._object_extractor = extractor
            return self

        def extract_objects(self, *args, **kwargs):
            if self._object_extractor is None:
                return None, None, None, False
            try:
                object_info = self._object_extractor(*args, **kwargs)
            except Exception as e:
                logger.warning("Error using object extractor: " + str(e))
                object_info = {'id': (-1), 'type': None}
            assert isinstance(object_info, dict), "Object extractors must return a dictionary."
            assert len(set(['id', 'type']).difference(object_info.keys())) == 0 and len(set(object_info.keys()).difference(['id', 'type', 'action', 'may_change'])) == 0, "Object extractors must at least include the keys 'id' and 'type', and optionally 'action' and 'may_change'. Was provided with: {}".format(str(list(object_info.keys())))
            object_info = defaultdict(lambda: None, object_info)
            return object_info['id'], object_info['type'], object_info['action'], object_info['may_change'] or False


class Vote(db.Model):
    __tablename__ = 'votes'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer)
    object_id = db.Column(db.Integer)
    object_type = db.Column(db.String(100), default='post')
    created_at = db.Column(db.DateTime, default=func.now())
    updated_at = db.Column(db.DateTime, default=func.now(), onupdate=func.now())


@unique_constructor(
    lambda username: username,
    lambda query, username: query.filter(User.username == username)
)
class User(db.Model):
    __tablename__ = 'users'

    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(500))
    created_at = db.Column(db.DateTime, default=func.now())

    _posts_assoc = db.relationship("PostAuthorAssoc")
    posts = association_proxy('_posts_assoc', 'post')  # This property should not directly modified

    @property
    def format_name(self):
        username_to_name = current_repo.config.username_to_name
        return username_to_name(self.username)

    @property
    def get_subscriptions(self):
        """Get the subscriptions associated with a user.

        Return an array of strings of tag_names
        """
        subscriptions = (db.session.query(Subscription)
                         .filter(Subscription.user_id == self.id)
                         .all())
        out_subscriptions = []
        for s in subscriptions:
            if s.object_type == 'tag':
                tag_obj = (db.session.query(Tag)
                           .filter(Tag.id == s.object_id)
                           .first())
                if tag_obj:
                    full_name = tag_obj.name
                    out_subscriptions.append(full_name)
                else:
                    db.session.delete(s)
        return out_subscriptions

    @property
    def get_liked_posts(self):
        """
        :return: Posts that a user has liked
        :rtype: list
        """
        votes = (db.session.query(Vote)
                 .filter(Vote.user_id == self.id)
                 .all())
        post_ids = [vote.object_id for vote in votes]
        posts = (db.session.query(Post)
                 .filter(Post.id.in_(post_ids))
                 .all())
        return posts


@unique_constructor(
    lambda name: name,
    lambda query, name: query.filter(Tag.name == name)
)
class Tag(db.Model):
    __tablename__ = 'tags'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(500))
    _description = db.Column('description', db.Text())
    created_at = db.Column(db.DateTime, default=func.now())

    @hybrid_property
    def description(self):
        if self._description:
            return self._description
        return "All posts with tag '{}'.".format(self.name)

    @description.expression
    def description(self):
        raise NotImplementedError


class Subscription(db.Model):
    __tablename__ = 'subscriptions'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer)
    object_id = db.Column(db.Integer)
    object_type = db.Column(db.String(100))  # Currently just tag
    created_at = db.Column(db.DateTime, default=func.now())


class Post(db.Model):
    __tablename__ = 'posts'

    id = db.Column(db.Integer, primary_key=True)
    uuid = db.Column(db.String(100), unique=True)
    path = db.Column(db.String(512), unique=True)
    project = db.Column(db.String(512), nullable=True)  # DEPRECATED
    repository = db.Column(db.String(512))
    revision = db.Column(db.Integer())

    title = db.Column(db.Text())
    tldr = db.Column(db.Text)
    keywords = db.Column(db.Text)
    thumbnail = db.Column(db.Text())

    created_at = db.Column(db.DateTime, default=func.now())
    updated_at = db.Column(db.DateTime, default=func.now())

    _authors_assoc = db.relationship("PostAuthorAssoc",
                                     order_by='PostAuthorAssoc.order',
                                     collection_class=ordering_list('order'),
                                     cascade="all, delete-orphan")
    _authors = association_proxy('_authors_assoc', 'author',
                                 creator=lambda author: PostAuthorAssoc(author=author),)

    @property
    def authors(self):
        return self._authors

    @authors.setter
    def authors(self, authors):
        """
        Sets the tags of the post to the tags given in comma delimited string
        form in tags_string
        """
        user_objs = []

        for author in authors:
            if not isinstance(author, User):
                author = author.strip()
                author = User(username=author)
            user_objs.append(author)

        self._authors = user_objs

    @hybrid_property
    def authors_string(self):
        return ', '.join([author.format_name for author in self.authors])

    @authors_string.expression
    def authors_string(self):
        raise NotImplementedError

    _tags = db.relationship("Tag", secondary=assoc_post_tag, backref='posts',
                            lazy='subquery')

    @hybrid_property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, tags):
        """
        Sets the tags of the post to the tags given in comma delimited string
        form in tags_string
        """
        tag_objs = []

        for tag in tags:
            if not isinstance(tag, Tag):
                tag = tag.strip()
                if tag[0] == "#":
                    tag = tag[1:]
                tag = Tag(name=tag)
            tag_objs.append(tag)

        self._tags = tag_objs

    _status = db.Column('status', db.Integer(), default=0)

    @hybrid_property
    def status(self):
        return current_repo.PostStatus(self._status or 0)

    @status.expression
    def status(self):
        return func.coalesce(self._status, 0)

    @status.setter
    def status(self, status):
        if status is None:
            self._status = None
        else:
            assert isinstance(status, KnowledgeRepository.PostStatus), "Status must be an instance of KnowledgeRepository.PostStatus.Status or None"
            self._status = status.value

    @hybrid_property
    def is_published(self):
        return self.status == current_repo.PostStatus.PUBLISHED

    @is_published.expression
    def is_published(self):
        return func.coalesce(self._status, 0) == current_repo.PostStatus.PUBLISHED.value

    _views = db.relationship("PageView", lazy='dynamic',
                             primaryjoin="and_(foreign(PageView.object_id)==Post.id, "
                                         "PageView.object_type=='post',"
                                         "PageView.object_action=='view')")

    @hybrid_property
    def views(self):
        return self._views.all()

    @hybrid_property
    def view_count(self):
        return self._views.count()

    @view_count.expression
    def view_count(self):
        return (select([func.count(PageView.id)])
                .where(PageView.object_id == self.id)
                .where(PageView.object_type == 'post')
                .label("view_count"))

    @hybrid_property
    def view_user_count(self):
        return (db.session.query(func.count(distinct(PageView.user_id)))
                          .filter(PageView.object_id == self.id)
                          .filter(PageView.object_type == 'post')
                          .scalar())

    @view_user_count.expression
    def view_user_count(self):
        return (select([func.count(distinct(PageView.user_id))])
                .where(PageView.object_id == self.id)
                .where(PageView.object_type == 'post')
                .label("view_user_count"))

    _votes = db.relationship("Vote", lazy='dynamic',
                             primaryjoin="and_(foreign(Vote.object_id)==Post.id, "
                                         "Vote.object_type=='post')")

    @hybrid_property
    def votes(self):
        return self._votes.all()

    @hybrid_property
    def vote_count(self):
        """ Given the path of a post, return the total likes """
        return self._votes.count()

    @vote_count.expression
    def vote_count(self):
        return (select([func.count(Vote.id)])
                .where(Vote.object_id == self.id)
                .where(Vote.object_type == 'post')
                .label("vote_count"))

    def vote_counted_for_user(self, user_id):
        return (db_session.query(Vote)
                          .filter(and_(Vote.object_id == self.id, Vote.object_type == 'post', Vote.user_id == user_id))
                          .first()) is not None

    _comments = db.relationship("Comment", lazy="dynamic",
                                primaryjoin="and_(foreign(Comment.post_id)==Post.id, "
                                            "Comment.type=='post')")

    @hybrid_property
    def comments(self):
        return self._comments.all()

    @hybrid_property
    def comment_count(self):
        """ Given the path of the a post, return the total comments """
        return self._comments.count()

    @comment_count.expression
    def comment_count(self):
        return (select([func.count(Comment.id)])
                .where(Comment.post_id == self.id)
                .where(Comment.object_type == 'post')
                .label("comments_count"))

    @property
    def kp(self):
        return current_repo.post(self.path)

    @property
    def text(self):
        return self.kp.read()

    def update_metadata_from_kp(self, kp):
        """
        :param kp: Maps fields of the model to values
        :type kp: KnowledgePost
        :param kp: Maps fields of the model to values
        :type kr: KnowledgeRepository
        :return: None
        :rtype: None
        """
        headers = kp.headers

        self.uuid = kp.uuid
        self.path = kp.path
        self.project = headers.get('project')
        self.repository = kp.repository_uri
        self.revision = kp.revision
        self.title = headers['title']
        self.tldr = headers['tldr']
        self.keywords = ",".join(headers.get('keywords', []))
        self.thumbnail = kp.thumbnail_uri

        self.created_at = headers['created_at']
        self.updated_at = headers['updated_at']
        if self.created_at > self.updated_at:
            self.updated_at = self.created_at

        self.authors = headers.get('authors', [])
        self.tags = headers.get('tags', [])
        self.status = kp.status


class Email(db.Model):

    __tablename__ = 'emails'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer)
    trigger_id = db.Column(db.Integer)
    trigger_type = db.Column(db.String(100))
    object_id = db.Column(db.Integer)
    object_type = db.Column(db.String(100))
    sent_at = db.Column(db.DateTime, default=func.now())
    subject = db.Column(db.Text)
    text = db.Column(db.Text)



from ..converter import KnowledgePostConverter


class StubConverter(KnowledgePostConverter):
    '''
    Use this as a template for new KnowledgePostConverters.
    '''
    _registry_keys = None

    @property
    def dependencies(self):
        # Dependencies required for this converter on top of core knowledge-repo dependencies
        return []

    def from_file(self, filename, **opts):
        raise NotImplementedError

    def from_string(self, filename, **opts):
        raise NotImplementedError

    def to_file(self, filename, **opts):
        raise NotImplementedError

    def to_string(self, **opts):
        raise NotImplementedError



import os
import logging
import subprocess
import tempfile

from ..converter import KnowledgePostConverter


logger = logging.getLogger(__name__)


class RmdConverter(KnowledgePostConverter):
    _registry_keys = ['rmd']

    def from_file(self, filename, rebuild=True):
        Rmd_filename = filename
        if rebuild:
            _, tmp_path = tempfile.mkstemp()

            runcmd = """R --vanilla --slave -e "library(knitr); setwd('{0}'); \
                        x = knit('{1}', '{2}', quiet=T)" """.format(os.path.abspath(os.path.dirname(filename)),
                                                                    os.path.abspath(filename),
                                                                    tmp_path)
            subprocess.check_output(runcmd, shell=True)
            Rmd_filename = tmp_path

        with open(Rmd_filename) as f:
            self.kp.write(f.read())
        self.kp.add_srcfile(filename)

        # Clean up temporary file
        if rebuild:
            os.remove(tmp_path)



import datetime
import json
import collections
from flask import Blueprint, request, render_template
from sqlalchemy import func

from ..app import db_session
from ..models import PageView, Post
from ..utils.requests import from_request_get_feed_params

blueprint = Blueprint('stats', __name__,
                      template_folder='../templates', static_folder='../static')


def _round_date(d):
    """Round date to the previous Monday"""
    return d - datetime.timedelta(d.weekday())


@blueprint.route('/post_stats', methods=['GET'])
@PageView.logged
def post_stats():
    """ Return the total views, distinct views, total likes and total comments
        for a given post """
    post_id = request.args.get('post_id', '')
    post = (db_session.query(Post)
                      .filter(Post.path == post_id)
                      .first())
    if not post:
        return json.dumps({})
    return json.dumps({'all_views': post.view_count,
                       'distinct_views': post.view_user_count,
                       'total_likes': post.vote_count,
                       'total_comments': post.comment_count})


@blueprint.route('/stats', methods=['GET'])
@PageView.logged
def stats():
    """ Render the stats page, creating graphs for
        pageviews daily, weekly, post creation weekly and
        cumulative posts per week
    """
    feed_params = from_request_get_feed_params(request)

    # count daily and weekly pageviews
    datetime_pageviews = (db_session.query(PageView.created_at, func.count(PageView.id))
                                    .group_by(PageView.created_at)
                                    .all())

    daily_pageviews = {}
    weekly_pageviews = {}
    for k, v in datetime_pageviews:
        d = k.date()
        wk = _round_date(d)
        daily_pageviews[d] = daily_pageviews.get(d, 0) + v
        weekly_pageviews[wk] = weekly_pageviews.get(wk, 0) + v

    # count post creation (created and updated)
    posts = (db_session.query(Post)
                       .filter(Post.is_published).all())

    created_at_counts = collections.Counter([_round_date(post.created_at.date()) for post in posts])
    updated_at_counts = collections.Counter([_round_date(post.updated_at.date()) for post in posts])

    all_week_keys = set(created_at_counts.keys()).union(updated_at_counts.keys())
    weekly_posts_created_and_updated = dict((k, [created_at_counts.get(k, 0), updated_at_counts.get(k, 0)])
                                            for k in all_week_keys)

    # cumulative weekly post created
    weekly_cumulative_posts = {}
    cum_created_val = 0
    for week in all_week_keys:
        cum_created_val += created_at_counts.get(week, 0)
        weekly_cumulative_posts[week] = cum_created_val

    return render_template('stats.html',
                           feed_params=feed_params,
                           daily_pageviews=daily_pageviews,
                           weekly_posts_created_and_updated=weekly_posts_created_and_updated,
                           weekly_cumulative_posts=weekly_cumulative_posts,
                           weekly_pageviews=weekly_pageviews)



import unittest

from knowledge_repo import KnowledgeRepository


class ClusterTest(unittest.TestCase):

    def setUp(self):
        self.repo = KnowledgeRepository.for_uri('tests/test_repo', auto_create=True)
        self.app = self.repo.get_app(config='tests/config_server.py').test_client()
        self.headers = {}
#         username_request_header = app.config.get("USERNAME_REQUEST_HEADER")
#         self.knowledge_username = 'clusters_test_user'
#         self.headers = {
#             username_request_header: self.knowledge_username
#         }
#         engine.echo = False

    def test01_test_cluster_folder(self):
        """
        test the cluster folder view
        """
        folder_alpha = '/cluster?group_by=folder&sort_by=alpha&sort_asc=1'
        folder_num = '/cluster?group_by=folder&sort_by=count'

        alpha_response = self.app.get(folder_alpha, headers=self.headers)
        assert (alpha_response.status == '200 OK')

        num_response = self.app.get(folder_num, headers=self.headers)
        assert (num_response.status == '200 OK')

    def test02_test_cluster_author(self):
        """
        test the cluster author view
        """
        author_alpha = '/cluster?group_by=author&sort_by=alphas&sort_asc=1'
        author_num = '/cluster?group_by=author&sort_by=count'

        alpha_response = self.app.get(author_alpha, headers=self.headers)
        assert (alpha_response.status == '200 OK')

        num_response = self.app.get(author_num, headers=self.headers)
        assert (num_response.status == '200 OK')

    def test03_test_cluster_tags(self):
        """
        test the cluster tags view
        """
        tags_alpha = '/cluster?group_by=tags&sort_by=alphas&sort_asc=1'
        tags_num = '/cluster?group_by=tags&sort_by=count'

        alpha_response = self.app.get(tags_alpha, headers=self.headers)
        assert (alpha_response.status == '200 OK')

        num_response = self.app.get(tags_num, headers=self.headers)
        assert (num_response.status == '200 OK')


if __name__ == '__main__':
    unittest.main()



"""assoc_rename_and_uuid_column

Revision ID: 1b385158fd32
Revises: 40dfdcb2b07c
Create Date: 2016-08-17 14:37:40.391822

"""

# revision identifiers, used by Alembic.
revision = '1b385158fd32'
down_revision = '40dfdcb2b07c'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.rename_table('knowledge_post_author', 'assoc_post_author')
    op.rename_table('knowledge_post_tags', 'assoc_post_tag')
    op.add_column('assoc_post_author', sa.Column('order', sa.Integer(), nullable=True))

    op.add_column('posts', sa.Column('uuid', sa.String(length=100), nullable=True))
    op.create_unique_constraint(None, 'posts', ['uuid'])

    op.add_column('pageviews', sa.Column('object_action', sa.String(length=100), nullable=True))
    op.add_column('pageviews', sa.Column('version', sa.String(length=100), nullable=True))


def downgrade():
    op.drop_column('assoc_post_author', 'order')
    op.rename_table('assoc_post_author', 'knowledge_post_author')
    op.rename_table('assoc_post_tag', 'knowledge_post_tags')

    op.drop_constraint(None, 'posts', type_='unique')
    op.drop_column('posts', 'uuid')

    op.drop_column('pageviews', 'object_action')
    op.drop_column('pageviews', 'version')



from builtins import object
import re


class SubstitutionMapper(object):

    def __init__(self, patterns={}, mappers=[]):
        self.mappers = mappers
        self.patterns = {name: re.compile(pattern)
                         for name, pattern in patterns.items()}

    def apply(self, text):
        matches = self.find_matches(text)
        output = []
        last_offset = 0

        for dmatch in matches:
            name, offset, match = dmatch['name'], dmatch[
                'offset'], dmatch['match']

            output.append(text[last_offset:offset])
            last_offset = offset + len(match.group(0))
            for mapper in self.mappers:
                replacement = mapper(name, match)
                if replacement is not None:
                    break
            if replacement is None:
                output.append(match.group(0))
            else:
                output.append(replacement)
        output.append(text[last_offset:])
        return ''.join(output)

    def find_matches(self, text, reverse=False):
        matches = []
        for name in self.patterns:
            matches.extend(self.find_matches_for_pattern(name, text))
        return sorted(matches, key=lambda x: x['offset'], reverse=reverse)

    def find_matches_for_pattern(self, name, text):
        return [{'name': name,
                 'offset': m.start(),
                 'match': m} for m in self.patterns[name].finditer(text)]



import unittest

from sqlalchemy import and_

from knowledge_repo import KnowledgeRepository
from knowledge_repo.app.models import Post, Subscription, Email, User
from knowledge_repo.app.utils.emails import send_internal_error_email, send_subscription_emails, send_comment_email, send_review_email
from knowledge_repo.app.proxies import db_session


class EmailTest(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        self.repo = KnowledgeRepository.for_uri('tests/test_repo', auto_create=True)
        self.app = self.repo.get_app(config='tests/config_server.py')
        self.app.repository.config.editors = ['knowledge_editors']
        self.client = self.app.test_client()

        self.knowledge_username = 'email_test_user'
        # this creates a user
        with self.app.app_context():
            User(username=self.knowledge_username)

            web_post = Post(title='test_webpost')
            git_post = (db_session.query(Post)
                        .first())

            db_session.add(web_post)
            web_post.authors = ['test_author']

            git_post.tags = ['test_tag']

            db_session.commit()

            self.web_post_id = web_post.id
            self.git_post_id = git_post.id

    def test01_review_email(self):
        """ Test Review Email """
        commenter = 'test_commenter'
        comment_text = 'this is a test for review email'
        with self.app.app_context():
            with self.app.config['mail'].record_messages() as outbox:
                with self.app.test_request_context():
                    send_review_email(commenter=commenter,
                                      comment_text=comment_text,
                                      post_id=self.web_post_id)

            web_post = (db_session.query(Post)
                        .filter(Post.id == self.web_post_id)
                        .first())
            assert len(outbox) == 1
            assert outbox[0].subject == "Someone reviewed your web post!"
            # TODO assert format_commenter(commenter) in outbox[0].body
            assert comment_text in outbox[0].body
            assert web_post.title in outbox[0].body

    def test02_comment_email(self):
        """ Test Comment Email """
        with self.app.app_context():
            with self.app.config['mail'].record_messages() as outbox:
                with self.app.test_request_context():
                    commenter = 'test_commenter'
                    comment_text = 'this is a test for comment email'
                    send_comment_email(commenter=commenter,
                                       comment_text=comment_text,
                                       post_id=self.git_post_id)

            git_post = (db_session.query(Post)
                        .filter(Post.id == self.git_post_id)
                        .first())
            assert len(outbox) == 1
            assert outbox[0].subject == "Someone commented on your post!"
            # assert format_commenter(commenter) in outbox[0].body
            assert comment_text in outbox[0].body
            assert git_post.title in outbox[0].body

    def test03_internal_error_email(self):
        """ Test Internal Error Email """
        with self.app.app_context():
            with self.app.config['mail'].record_messages() as outbox:
                with self.app.app_context():
                    subject_line = 'test subject line'
                    send_internal_error_email(subject_line=subject_line)

                    assert len(outbox) == 1
                    assert outbox[0].subject == subject_line

                    # make sure that internal emails are sent to the editors
                    # TODO does this make sense for open source?
                    username_to_email = self.app.repository.config.username_to_email
                    knowledge_editors = self.app.repository.config.editors
                    editor_emails = [username_to_email(
                        editor) for editor in knowledge_editors]
                    assert sorted(outbox[0].recipients) == sorted(editor_emails)
                    assert 'This email is for internal debugging only.' in outbox[
                        0].body

    def test04_subscription_email(self):
        """ Test Subscription Email """
        # Subscribe to a tag from gitpost and webpost Note, don't send
        # a request to do it so that create_index() will happen later,
        # as it is before_first_request
        with self.app.app_context():
            db_session.expire_on_commit = False
            git_post = (db_session.query(Post)
                        .filter(Post.id == self.git_post_id)
                        .first())

            user = User(username=self.knowledge_username)
            if user.id is None:
                db_session.commit()
            user_id = user.id

            # grab the first tag from the post
            tag_id = git_post.tags[0].id
            post_id = git_post.id

            subscription = Subscription(user_id=user_id, object_type='tag',
                                        object_id=tag_id)
            db_session.add(subscription)
            db_session.commit()

            # Run a request, which should trigger a subscription email for
            # tag_to_subscribe
            with self.app.config['mail'].record_messages() as outbox:
                with self.app.app_context():
                    send_subscription_emails(git_post)

            emails_sent = (db_session.query(Email)
                           .filter(and_(Email.object_id == post_id,
                                        Email.user_id == user_id,
                                        Email.trigger_id == tag_id,
                                        Email.trigger_type == 'subscription'))
                           .all())

            # There should be exactly 1 email sent to the user for the target post
            # Check EmailSent records
            assert len(outbox) == 1, 'One subscription email should be actually sent'
            assert len(
                emails_sent) == 1, 'One subscription email should recorded as sent'
            # Check outbox of messages
            assert len(
                outbox[0].bcc) == 1, 'One subscription email should be actually sent to one user'
            assert "A knowledge post was just posted under tag" in outbox[0].body, 'Email should be the subscription email'

            username_to_email = self.app.repository.config.username_to_email
            assert outbox[0].bcc[0] == username_to_email(self.knowledge_username)

if __name__ == '__main__':
    unittest.main()



from builtins import range
from builtins import str
import json
import logging
import sys
import os
from collections import defaultdict
from datetime import datetime
from flask import request, render_template, Blueprint, current_app, url_for, send_from_directory, g
from sqlalchemy import or_
from werkzeug import secure_filename

from knowledge_repo.post import KnowledgePost
from ..proxies import db_session, current_repo
from ..models import Post, PostAuthorAssoc, Tag, Comment, User, PageView
from ..utils.emails import send_review_email, send_reviewer_request_email
from ..utils.image import pdf_page_to_png, is_pdf, is_allowed_image_format


if sys.version_info.major > 2:
    from urllib.parse import unquote as urlunquote
else:
    from urllib import unquote as urlunquote

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


blueprint = Blueprint('web_editor', __name__,
                      template_folder='../templates', static_folder='../static')

# TODO: These functions have not been fully married to the KnowledgePost API
# Currently, backended by Post objects but partially implemented on KnowledgePost API


@blueprint.route('/ajax_tags_typeahead', methods=['GET'])
def generate_tags_typeahead():
    tag_objs = (db_session.query(Tag).all())
    tags_typeahead = [str(tag_obj.name) for tag_obj in tag_objs]
    return json.dumps(tags_typeahead)


@blueprint.route('/ajax_users_typeahead', methods=['GET'])
def generate_users_typeahead():
    user_objs = (db_session.query(User).all())
    users_typeahead = [str(user_obj.username) for user_obj in user_objs]
    return json.dumps(users_typeahead)


@blueprint.route('/ajax_projects_typeahead', methods=['GET'])
def generate_projects_typeahead():
    posts = db_session.query(Post).all()
    projects = []
    for p in posts:
        if p.project:
            projects.append(p.project)
        elif p.path:
            projects.append(from_path_get_project(p.path))
    return json.dumps(list(set(projects)))


def from_path_get_project(path):
    """
    To maintain an organization of webposts beyond a top-level
    grouping, we keep track of a "project" for each post
    :param: path of the post
    :type path: string
    """
    folder_list = path.split("/")
    # All posts are under the "projects/" folder
    # We assume the next folder is the project
    if len(folder_list) >= 2:
        return folder_list[1]
    return ""


@blueprint.route('/webposts', methods=['GET'])
@PageView.logged
def gitless_drafts():
    """ Render the gitless posts that a user has created in table form
        Editors can see all the posts created via Gitless_Editing
    """
    query = (db_session.query(Post))
    if current_app.config.get('WEB_EDITOR_PREFIXES', None):
        query = query.filter(or_(*[Post.path.like(p + '%') for p in current_app.config['WEB_EDITOR_PREFIXES']]))

    if g.user.username not in current_repo.config.editors:
        query = (query.outerjoin(PostAuthorAssoc, PostAuthorAssoc.post_id == Post.id)
                      .filter(PostAuthorAssoc.user_id == g.user.id))

    webposts = query.all()

    return render_template("web_posts.html", posts=webposts)


@blueprint.route('/posteditor', methods=['GET', 'POST'])
@PageView.logged
def post_editor():
    """ Render the web post editor, either with the default settings
        or if the post already exists, with what has been saved
    """
    post_id = request.args.get('post_id', None)

    headers = defaultdict(lambda: None)
    status = current_repo.PostStatus.DRAFT
    markdown = None
    comments = None
    thumbnail = ''
    if post_id and post_id != 'None':
        webpost = (db_session.query(Post)
                             .filter(Post.id == post_id)
                             .first())
        webpost_path = webpost.path
        thumbnail = webpost.thumbnail
        kp = current_repo.post(webpost_path)
        status = current_repo.post_status(webpost_path)
        headers = kp.headers
        markdown = kp.read(images=False, headers=False)
        comments = (db_session.query(Comment)
                              .filter(Comment.post_id == post_id)
                              .filter(Comment.type == "review")
                              .all())

    can_approve = 0
    authors_str = ''
    if 'authors' in headers:
        authors_str = ', '.join(headers['authors'])
        if g.user.username not in authors_str:
            can_approve = 1

    tags_str = ''
    if 'tags' in headers:
        tags_str = ', '.join(headers['tags'])

    # get created at and updated at, or default to now
    created_at = (headers['created_at'] or datetime.now()).strftime('%Y-%m-%d')
    updated_at = (headers['updated_at'] or datetime.now()).strftime('%Y-%m-%d')

    return render_template('post_editor.html',
                           title=headers['title'],
                           project=headers.get('project', ''),
                           created_at=created_at,
                           updated_at=updated_at,
                           author=authors_str,
                           tags=tags_str,
                           tldr=headers['tldr'],
                           markdown=markdown,
                           post_id=post_id,
                           status=status.value,
                           comments=comments,
                           feed_image=thumbnail,
                           can_approve=can_approve,
                           username=g.user.username)


@blueprint.route('/submit', methods=['GET', 'POST'])
@PageView.logged
def submit_for_review():
    """ Change the status of a gitless post
        And if there are reviewers assigned, email them
    """
    post_id = request.args.get('post_id', None)
    data = request.get_json()
    post = (db_session.query(Post)
                      .filter(Post.id == post_id)
                      .first())
    post.status = current_repo.PostStatus.SUBMITTED
    db_session.commit()

    path = post.path
    current_repo.submit(path)

    # email the reviewers
    if not data['post_reviewers']:
        return
    reviewers = data['post_reviewers'].split(",")
    if reviewers:
        for r in reviewers:
            send_reviewer_request_email(post_id=post_id, reviewer=r)
    return ""


@blueprint.route('/publish_post', methods=['GET', 'POST'])
@PageView.logged
def publish_post():
    """ Publish the gitless post by changing the status """
    post_id = request.args.get('post_id', None)
    post = (db_session.query(Post)
                      .filter(Post.id == post_id)
                      .first())
    if not post:
        error_msg = "Unable to retrieve post with id = {}!".format(str(post_id))
        json_str = json.dumps({'msg': error_msg, 'success': False})
        return json_str

    post.status = current_repo.PostStatus.PUBLISHED
    db_session.commit()

    path = post.path
    current_repo.publish(path)
    return ""


@blueprint.route('/unpublish_post', methods=['GET', 'POST'])
@PageView.logged
def unpublish_post():
    """ Unpublish the post, going back to the in_review status """
    post_id = request.args.get('post_id', None)
    post = (db_session.query(Post)
                      .filter(Post.id == post_id)
                      .first())
    post.status = current_repo.PostStatus.UNPUBLISHED
    db_session.commit()

    path = post.path
    current_repo.unpublish(path)
    return ""


@blueprint.route('/author_publish', methods=['POST'])
@PageView.logged
def change_author_publish():
    """ Allow the author to publish their own webpost if the
        editor allows them to
    """
    post_id = request.args.get('post_id', None)
    post = (db_session.query(Post)
                      .filter(Post.id == post_id)
                      .first())
    post.status = current_repo.PostStatus.UNPUBLISHED
    db_session.commit()

    path = post.path
    current_repo.accept(path)
    return ""


@blueprint.route('/save_post', methods=['GET', 'POST'])
@PageView.logged
def save_post():
    """ Save the gitless post """
    post_id = request.args.get('post_id', None)
    data = request.get_json()
    post = (db_session.query(Post)
            .filter(Post.id == post_id)
            .first())
    new_post = False
    if not post:
        new_post = True
        path = "{}/{}.kp".format(data['project'],
                                 data['title'].encode('utf8').lower().replace(' ', '_'))
        if current_app.config.get('WEB_EDITOR_PREFIXES', None):
            # TODO: Include dropdown on webeditor to have user specify repo
            path = "{}/{}".format(current_app.config['WEB_EDITOR_PREFIXES'][0], path)

        post = (db_session.query(Post)
                          .filter(Post.path == path)
                          .first())
        if post:
            error_msg = "Post with project {} and title {} already exists!".format(data['project'], data['title'])

            json_str = json.dumps({'msg': error_msg, 'success': False})
            return json_str
        else:
            post = Post()
            post.path = path
    else:
        path = post.path

    # create the knowledge post
    kp = KnowledgePost(path=path)

    headers = {}
    headers['created_at'] = datetime.strptime(data['created_at'], '%Y-%m-%d')
    headers['updated_at'] = datetime.strptime(data['updated_at'], '%Y-%m-%d')
    headers['title'] = str(data['title'])
    headers['path'] = str(post.path)
    headers['project'] = str(data['project'])
    # TODO: thumbnail header not working currently, as feed image set with kp
    # method not based on header
    headers['thumbnail'] = str(data['feed_image'])
    headers['authors'] = [str(auth).strip() for auth in data['author']]
    headers['tldr'] = str(data['tldr'])
    headers['tags'] = [str(tag).strip() for tag in data['tags']]
    kp.write(urlunquote(str(data['markdown'])), headers=headers)

    # add to repo
    current_repo.add(kp, update=True)  # THIS IS DANGEROUS
    db_session.commit()

    # add to index
    post.update_metadata_from_kp(kp)
    if new_post:
        db_session.add(post)
    db_session.commit()

    return json.dumps({'post_id': str(post.id)})


@blueprint.route('/delete_post', methods=['GET'])
@PageView.logged
def delete_post():
    """ Delete a post """
    post_id = request.args.get('post_id', None)
    item = (db_session.query(Post)
                      .filter(Post.id == post_id)
                      .first())
    db_session.delete(item)
    db_session.commit()
    return ""


@blueprint.route('/file_upload', methods=['POST', 'GET'])
@PageView.logged
def file_upload():
    """ Uploads images dropped on the web editor's markdown box to static/images
        and notifies editors by email
    """
    upload_folder = 'images'
    title = request.form['title']
    files = request.files
    uploadedFiles = []

    if files:
        for img_file in files.values():
            filename = secure_filename(title + "_" + img_file.filename).lower()
            dst_folder = os.path.join(current_app.static_folder, upload_folder)

            if is_allowed_image_format(img_file):
                try:
                    img_file.save(os.path.join(dst_folder, filename))
                    send_from_directory(dst_folder, filename)
                    uploadedFiles += [url_for("static", filename=os.path.join(upload_folder, filename))]
                except Exception as e:
                    error_msg = "ERROR during image upload: {}".format(str(e))
                    logger.error(error_msg)
                    return json.dumps({'error_msg': error_msg, 'success': False})

            elif is_pdf(filename):
                from PyPDF2 import PdfFileReader
                try:
                    src_pdf = PdfFileReader(img_file)
                    filename = os.path.splitext(filename)[0]
                    num_pages = src_pdf.getNumPages()
                    for page_num in range(num_pages):
                        page_png = pdf_page_to_png(src_pdf, page_num)
                        page_name = "{filename}_{page_num}.jpg".format(**locals())
                        page_png.save(filename=os.path.join(dst_folder, page_name))
                        uploadedFiles += [url_for("static", filename=os.path.join(upload_folder, page_name))]
                except Exception as e:
                    error_msg = "ERROR during pdf upload: {}".format(str(e))
                    logger.error(error_msg)
                    return json.dumps({'error_msg': error_msg, 'success': False})

    return json.dumps({'links': uploadedFiles, 'success': True})


@blueprint.route('/review', methods=['POST'])
@PageView.logged
def review_comment():
    """ Saves the comments underneath a Gitless Post to a table
        and sends an email that the post has been reviewed to the
        author of the post
    """
    post_id = request.args.get('post_id', None)
    commenter, commenter_id = g.user.username, g.user.id
    data = request.get_json()   # just the comment text
    comment = Comment()
    comment.text = data['text']
    comment.user_id = commenter_id
    comment.post_id = post_id
    comment.type = "review"
    db_session.add(comment)
    db_session.commit()

    send_review_email(post_id=post_id, commenter=commenter,
                      comment_text=comment.text)
    return ""


@blueprint.route('/delete_review', methods=['GET', 'POST'])
@PageView.logged
def delete_review():
    """ Delete a comment that was made underneath a gitless contribution post """
    try:
        comment_id = int(request.args.get('comment_id', ''))
        items = db_session.query(Comment).filter(
            Comment.id == comment_id).all()
        for item in items:
            db_session.delete(item)
        db_session.commit()
    except:
        logging.warning("ERROR processing request")
        pass
    return ""





from __future__ import absolute_import
from __future__ import unicode_literals
from markdown import Extension
from markdown.preprocessors import Preprocessor


class KnowledgeMetaExtension(Extension):
    """ Meta-Data extension for Python-Markdown. """

    def extendMarkdown(self, md, md_globals):
        """ Add MetaPreprocessor to Markdown instance. """
        md.preprocessors.add("knowledge_meta",
                             KnowledgeMetaPreprocessor(md),
                             ">normalize_whitespace")


class KnowledgeMetaPreprocessor(Preprocessor):
    """ Get Meta-Data. """

    def run(self, lines):
        """ Parse Meta-Data and store in Markdown.Meta. """
        cnt = 0
        for i, line in enumerate(lines):
            if line.strip() == '---':
                cnt = cnt + 1
            if cnt == 2:
                break
        return lines[i + 1:]


def makeExtension(*args, **kwargs):
    return KnowledgeMetaExtension(*args, **kwargs)



from builtins import object

import logging
import posixpath

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session, mapper
from sqlalchemy.sql import func

from sqlalchemy import Table, Column, Integer, String, DateTime, BLOB, MetaData
from ..repository import KnowledgeRepository

logger = logging.getLogger(__name__)


class DbKnowledgeRepository(KnowledgeRepository):
    _registry_keys = ['postgresql',
                      'postgresql+psycopg2',
                      'postgresql+pg8000',
                      'mysql',
                      'mysql+mysqldb',
                      'mysql+mysqlconnector',
                      'mysql+oursql',
                      'oracle',
                      'oracle+cx_oracle',
                      'mssql',
                      'mssql+pyodbc',
                      'mssql+pymssql',
                      'sqlite']  # TODO: Use regex matching and add the rest?

    def init(self, auto_create=True):

        # TODO handle if user does not pass in table sqlite://path.db
        uri_splt = self.uri.split(":")
        engine_uri = ":".join(uri_splt[:-1])
        table_name = uri_splt[-1]

        metadata = MetaData()
        postref_table = Table(table_name, metadata,
                              Column('id', Integer, primary_key=True),
                              Column('created_at', DateTime, default=func.now()),
                              Column('uuid', String(512)),
                              Column('path', String(512)),
                              Column('revision', Integer, default=0),
                              Column('status', Integer, default=self.PostStatus.DRAFT.value),
                              Column('ref', String(512)),
                              Column('data', BLOB))
        self.engine = create_engine(engine_uri)
        self.session = scoped_session(sessionmaker(bind=self.engine))
        if auto_create:
            postref_table.create(self.engine, checkfirst=True)

        class PostRef(object):
            pass
        mapper(PostRef, postref_table)
        self.PostRef = PostRef

    # ------------- Repository actions / state ------------------------------------
    def session_begin(self):
        pass

    def session_end(self):
        self.session.remove()

    @property
    def revision(self):
        return self.session.query(func.max(self.PostRef.created_at))

    def update(self):
        pass

    def set_active_draft(self, path):
        pass

    @property
    def status(self):
        try:
            session = self.Session()
            session.remove()
        except:
            return {'status': 'unable to connect to uri {}'.format(self.uri)}
        return {'status': 'super'}

    @property
    def status_message(self):
        return self.status['status']

    # -------------- Post retrieval methods --------------------------------------

    def _dir(self, prefix, statuses):
        query = self.session.query(self.PostRef.path)
        if prefix is not None:
            query = query.filter(self.PostRef.status.like('{}%'.format(prefix)))
        query = query.filter(self.PostRef.status.in_([status.value for status in statuses]))

        for (prefix,) in query.distinct():
            if prefix is not None:
                yield prefix

    # -------------- Post submission / addition user flow --------------------

    def _add_prepare(self, kp, path, update=False):
        pass

    def _add_cleanup(self, kp, path, update=False):
        self.__set_post_status(path, self.PostStatus.DRAFT, kp.revision)

    def _submit(self, path):  # Submit a post for review
        # Latest revision for draft
        revision = self._kp_get_revision(path, self.PostStatus.DRAFT, enforce_exists=True)
        self.__set_post_status(path, status=self.PostStatus.SUBMITTED, revision=revision)

    def _accept(self, path):  # Approve a post to publish
        # Latest revision for submitted draft
        revision = self._kp_get_revision(path, self.PostStatus.SUBMITTED, enforce_exists=True)
        if self.__get_post_status(path, revision) != self.PostStatus.PUBLISHED:
            self.__set_post_status(path, status=self.PostStatus.UNPUBLISHED, revision=revision)

    def _publish(self, path):  # Publish a post for general perusal
        # Latest revision for unpublished
        revision = self._kp_get_revision(path, self.PostStatus.UNPUBLISHED, enforce_exists=True)
        if self.__get_post_status(path, revision) == self.PostStatus.UNPUBLISHED:
            self.__set_post_status(path, status=self.PostStatus.PUBLISHED, revision=revision)

    def _unpublish(self, path):  # Unpublish a post
        # Latest revision for published post
        revision = self._kp_get_revision(path, self.PostStatus.PUBLISHED, enforce_exists=True)
        self.__set_post_status(path, status=self.PostStatus.UNPUBLISHED, revision=revision)

    def _remove(self, path, all=False):
        raise NotImplementedError

    def __set_post_status(self, path, status, revision=None):
        revision = revision or self._kp_get_revision(path, enforce_exists=True)
        postrefs = (self.session.query(self.PostRef)
                                .filter(self.PostRef.path == path)
                                .filter(self.PostRef.revision == revision)).all()
        for postref in postrefs:
            postref.status = status.value
        self.session.commit()

    def __get_post_status(self, path, revision=None):
        revision = revision or self._kp_get_revision(path, enforce_exists=True)
        post_refs = (self.session.query(self.PostRef)
                                 .filter(self.PostRef.path == path)
                                 .filter(self.PostRef.revision == revision)
                                 .distinct()).all()
        return self.PostStatus(min([pr.status if pr.status else 0 for pr in post_refs]))  # return the most 'open' status

    # ----------- Knowledge Post Data Retrieval/Pushing Methods --------------------

    def _kp_uuid(self, path):
        result = self.session.query(self.PostRef.uuid).filter(self.PostRef.path == path).first()
        if result:
            return result[0]
        return None

    def _kp_exists(self, path, revision=None):
        query = (self.session.query(self.PostRef)
                             .filter(self.PostRef.path == path))
        if revision:
            query = query.filter(self.PostRef.revision == revision)
        postref = query.first()
        return postref

    def _kp_status(self, path, revision=None, detailed=False):
        revision = revision or self._kp_get_revision(path, enforce_exists=True)

        postref = (self.session.query(self.PostRef)
                               .filter(self.PostRef.path == path)
                               .filter(self.PostRef.revision == revision)).first()

        if not postref:
            raise ValueError("No such post exists for path '{}' and revision '{}'.".format(path, revision))

        if detailed:
            return self.__get_post_status(path, revision), None
        return self.__get_post_status(path, revision)

    def _kp_get_revision(self, path, status=None, enforce_exists=False):
        query = self.session.query(self.PostRef.revision).filter(self.PostRef.path == path)
        if status is not None:
            query.filter(self.PostRef.status == status.value)
        revision = query.order_by(self.PostRef.revision.desc()).first()

        # TODO fix this
        if revision:
            revision = revision[0]
        if enforce_exists and revision is None:
            raise ValueError('No post found at {} (with status of {})'.format(path, status))
        return revision

    def _kp_get_revisions(self, path):
        revisions = (self.session.query(self.PostRef.revision)
                                 .filter(self.PostRef.path == path)
                                 .distinct()).all()
        return revisions

    def _kp_read_ref(self, path, reference, revision=None):
        revision = revision or self._kp_get_revision(path, enforce_exists=True)
        data = (self.session.query(self.PostRef)
                            .filter(self.PostRef.path == path)
                            .filter(self.PostRef.ref == reference)
                            .filter(self.PostRef.revision == revision)).first().data
        return data

    def _kp_dir(self, path, parent=None, revision=None):
        if parent:
            path = posixpath.join(path, parent)
        revision = revision or self._kp_get_revision(path, enforce_exists=True)
        refs = (self.session.query(self.PostRef.ref)
                            .filter(self.PostRef.path == path)
                            .filter(self.PostRef.revision == revision)).all()
        for (ref,) in refs:
            if ref is not None:
                yield ref

    def _kp_has_ref(self, path, reference, revision=None):
        revision = revision or self._kp_get_revision(path, enforce_exists=True)
        ref = (self.session.query(self.PostRef.ref)
                           .filter(self.PostRef.path == path)
                           .filter(self.PostRef.ref == reference)
                           .filter(self.PostRef.revision == revision)).first()

        return ref is not None

    def _kp_diff(self, path, head, base):
        raise NotImplementedError

    def _kp_write_ref(self, path, reference, data, uuid=None, revision=None):
        revision = revision or self._kp_get_revision(path, enforce_exists=False) or 0

        postref = (self.session.query(self.PostRef)
                               .filter(self.PostRef.path == path)
                               .filter(self.PostRef.ref == reference)
                               .filter(self.PostRef.revision == revision)).first()

        if postref is None:
            postref = self.PostRef()
            postref.path = path
            postref.ref = reference
            postref.uuid = uuid
            postref.revision = revision
            self.session.add(postref)

        postref.data = data
        self.session.commit()

    def _kp_new_revision(self, path, uuid=None):
        revision = self._kp_get_revision(path, enforce_exists=False) or 0
        return revision + 1



from __future__ import print_function
import unittest
import json

from bs4 import BeautifulSoup

from knowledge_repo import KnowledgeRepository
from knowledge_repo.app.models import Comment, Post, User

from knowledge_repo.app.app import db_session


class CommentTest(unittest.TestCase):

    def setUp(self):
        self.repo = KnowledgeRepository.for_uri('tests/test_repo', auto_create=True)
        self.repo_app = self.repo.get_app(config='tests/config_server.py')
        self.app = self.repo_app.test_client()

        self.knowledge_username = 'comment_test_user'
        with self.repo_app.app_context():
            user = User(username=self.knowledge_username)
            if user.id is None:
                db_session.commit()
            self.user_id = user.id
        username_request_header = self.repo_app.config.get(
            "AUTH_USERNAME_REQUEST_HEADER")
        self.headers = {
            username_request_header: self.knowledge_username
        }

    def test01_comment_on_a_post(self):
        """
        Comment on a post
        """
        with self.repo_app.app_context():
            # get the number of comments on a given post
            post = (db_session.query(Post)
                              .filter(Post.is_published)
                              .first())
            n_comments = len(db_session.query(Comment)
                                       .filter(Comment.post_id == post.id)
                                       .all())

            assert post.status == self.repo.PostStatus.PUBLISHED

            # test commenting on a post
            data = {'text': 'This is a comment'}
            rv = self.app.post('/comment?path={}'.format(post.path),
                               headers=self.headers,
                               data=json.dumps(data),
                               content_type='application/json')

            assert rv.status == '200 OK'

            assert post.status == self.repo.PostStatus.PUBLISHED

            # check that new comment exists in db
            comment_exists = (db_session.query(Comment)
                                        .filter(Comment.post_id == post.id)
                                        .filter(Comment.user_id == self.user_id)
                                        .filter(Comment.text == data['text'])
                                        .first())
            assert comment_exists

            assert post.status == self.repo.PostStatus.PUBLISHED

            # check that the number of comments rendered increased by 1
            rv = self.app.get('/render?markdown=' + post.path,
                              headers=self.headers)
            soup = BeautifulSoup(rv.data.decode('utf-8'), 'html.parser')
            comments = soup.findAll("div", {"class": "post_comment"})
            assert comments, 'comments found in '
            assert len(comments) == n_comments + 1

    def test02_delete_a_comment(self):
        """
        Delete a comment
        """

        with self.repo_app.app_context():

            # get a post
            post = (db_session.query(Post)
                              .filter(Post.is_published)
                              .first())

            # create a comment
            data = {'text': 'This is a comment'}
            rv = self.app.post('/comment?path={}'.format(post.path),
                               headers=self.headers,
                               data=json.dumps(data),
                               content_type='application/json')

            # get comment id
            comment_id = (db_session.query(Comment)
                                    .filter(Comment.post_id == post.id)
                                    .filter(Comment.user_id == self.user_id)
                                    .filter(Comment.text == data['text'])
                                    .first()).id

            # test deleting a commment
            rv = self.app.get('/delete_comment?comment_id={}'.format(comment_id),
                              headers=self.headers)
            assert rv.status == '200 OK'
            delete_comment = (db_session.query(Comment)
                                        .filter(Comment.id == comment_id)
                                        .all())
            assert not delete_comment, 'comments found after deletion'


if __name__ == '__main__':
    unittest.main()



import json
import logging

from flask import request, url_for, redirect, render_template, current_app, Blueprint, g

from ..proxies import db_session, current_repo
from ..models import User, Post, PageView
from ..utils.render import render_post, render_comment, render_post_raw


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


blueprint = Blueprint('render', __name__,
                      template_folder='../templates', static_folder='../static')


@blueprint.route('/raw', methods=['GET'])
@PageView.logged
def raw():
    """ Show the raw markdown of a post """
    return redirect(url_for('render.render', raw='true', **request.args))


@blueprint.route('/presentation', methods=['GET'])
@PageView.logged
def presentation():
    """ Render the knowledge post as a presentation """
    return redirect(url_for('render.render', presentation='true', **request.args))


@blueprint.route('/render', methods=['GET'])
@PageView.logged
def render():
    """ Render the knowledge post with all the related formatting """

    path = request.args.get('markdown', '')
    raw = request.args.get('raw', False)

    username, user_id = g.user.username, g.user.id

    tmpl = 'markdown-rendered.html'
    if raw:
        tmpl = 'markdown-raw.html'
    elif request.args.get('presentation'):
        # TODO(dan?) fix presentation post
        # presentation_post = {}
        # presentation_post['authors_string'] = post.author_string
        # presentation_post['tldr'] = post.tldr
        # presentation_post['html'] = html
        # html = create_presentation_text(presentation_post)
        tmpl = "markdown-presentation.html"

    if not current_app.config.get('REPOSITORY_INDEXING_ENABLED', True):
        return _render_preview(path=path, tmpl=tmpl)

    post = (db_session.query(Post)
                      .filter(Post.path == path)
                      .first())
    if not post:
        knowledge_aliases = current_repo.config.aliases
        if path in knowledge_aliases:
            # TODO: reframe as redirect
            post = (db_session.query(Post)
                              .filter(Post.path == knowledge_aliases[path])
                              .first())
    if not post:
        raise Exception("unable to find post at {}".format(path))

    html = render_post(post)
    raw_post = render_post_raw(post) if raw else None

    comments = post.comments
    for comment in comments:
        comment.author = db_session.query(User).filter(User.id == comment.user_id).first().username
        if not raw:
            comment.text = render_comment(comment)

    user_obj = (db_session.query(User)
                          .filter(User.id == user_id)
                          .first())

    tags_list = [str(t.name) for t in post.tags]
    user_subscriptions = [str(s) for s in user_obj.get_subscriptions]

    rendered = render_template(tmpl,
                               html=html,
                               post_id=post.id,
                               post_path=path,
                               raw_post=raw_post,
                               comments=comments,
                               username=username,
                               post_author=post.authors_string,
                               title=post.title,
                               page_views=post.view_count,
                               unique_views=post.view_user_count,
                               likes=post.vote_counted_for_user(user_id=user_id),
                               total_likes=post.vote_count,
                               tags_list=tags_list,
                               user_subscriptions=user_subscriptions,
                               webeditor_buttons=False,
                               web_uri=post.kp.web_uri,
                               table_id=None)
    return rendered


def _render_preview(path, tmpl):
    post = None
    if current_repo.has_post(path):
        post = current_repo.post(path)
    else:
        knowledge_aliases = current_repo.config.aliases
        if path in knowledge_aliases:
            # TODO: reframe as redirect
            post = current_repo.post(knowledge_aliases[path])

    if not post:
        raise Exception("unable to find post at {}".format(path))

    html = render_post(post)
    raw_post = render_post_raw(post) if raw else None

    return render_template(tmpl,
                           html=html,
                           post_id=None,
                           post_path=path,
                           raw_post=raw_post,
                           comments=[],
                           username=None,
                           post_author=', '.join(post.headers['authors']),
                           title=post.headers['title'],
                           page_views=0,
                           unique_views=0,
                           likes=1,
                           total_likes=0,
                           tags_list=','.join(post.headers.get('tags', [])),
                           user_subscriptions=[],
                           webeditor_buttons=False,
                           table_id=None)


@render.object_extractor
def render():
    post_path = request.args.get('markdown', '')
    return {
        'id': Post.query.filter(Post.path == post_path).first().id,
        'type': 'post',
        'action': 'view'
    }


@blueprint.route('/about', methods=['GET'])
@PageView.logged
def about():
    """Renders about page. This is the html version of REAMDE.md"""
    return render_template("about.html")


@blueprint.route('/ajax_post_typeahead', methods=['GET', 'POST'])
def ajax_post_typehead():
    return json.dumps(current_app.config.get('typeahead_data', {}))



from __future__ import absolute_import
import logging
from .proxies import db_session, current_repo
from .models import Post
from .utils.emails import send_subscription_emails

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def update_index():
    """ Initialize the db from a KnowledgeRepository object
    """
    kr_dir = {kp.path: kp for kp in current_repo.posts()}
    kr_uuids = {kp.uuid: kp for kp in kr_dir.values()}
    posts = db_session.query(Post).all()

    for post in posts:

        # If UUID has changed, check if we can find it elsewhere in the repository, and if so update index path
        if post.uuid and ((post.path not in kr_dir) or (post.uuid != kr_dir[post.path].uuid)):
            if post.uuid in kr_uuids:
                logger.info('Updating location of post: {} -> {}'.format(post.path, kr_uuids[post.uuid].path))
                post.path = kr_uuids[post.uuid].path

        # If path of post no longer in directory, mark as unpublished
        if post.path not in kr_dir:
            logger.info('Recording unpublished status for post at {}'.format(post.path))
            post.status = current_repo.PostStatus.UNPUBLISHED
            continue

        # Update database according to current state of existing knowledge post and
        # remove from kp_dir. This means that when this loop finishes, kr_dir will
        # only contain posts which are new to the repo.
        kp = kr_dir.pop(post.path)

        # Update metadata of post if required
        if (kp.revision > post.revision or not post.is_published):
            if kp.is_valid():
                logger.info('Recording update to post at: {}'.format(kp.path))
                post.update_metadata_from_kp(kp)
            else:
                logger.warning('Update to post at "{}" is corrupt.'.format(kp.path))

    # Add the new posts that remain in kr_dir
    for kp_path, kp in kr_dir.items():
        if not kp.is_valid():
            logger.warning('New post at "{}" is corrupt.'.format(kp.path))
            continue
        logger.info('creating new post from path {}'.format(kp_path))
        post = Post()
        db_session.add(post)
        post.update_metadata_from_kp(kp)
        send_subscription_emails(post)

    db_session.commit()



from ..repository import KnowledgeRepository


class StubKnowledgeRepository(KnowledgeRepository):
    _registry_keys = None

    def init(self):
        pass

    # ------------- Repository actions / state ------------------------------------

    def session_begin(self):
        pass

    def session_end(self):
        pass

    @property
    def revision(self):
        raise NotImplementedError

    def update(self):
        pass

    def set_active_draft(self, path):
        pass

    @property
    def status(self):
        raise NotImplementedError

    @property
    def status_message(self):
        raise NotImplementedError

    # -------------- Post retrieval methods --------------------------------------

    def _dir(self, prefix, statuses):
        raise NotImplementedError

    # -------------- Post submission / addition user flow --------------------

    def _add_prepare(self, kp, path, update=False):
        raise NotImplementedError

    def _add_cleanup(self, kp, path, update=False):
        raise NotImplementedError

    def _submit(self, path):  # Submit a post for review
        raise NotImplementedError

    def _accept(self, path):
        raise NotImplementedError

    def _publish(self, path):  # Publish a post for general perusal
        raise NotImplementedError

    def _unpublish(self, path):
        raise NotImplementedError

    def _remove(self, path, all=False):
        raise NotImplementedError

    # ----------- Knowledge Post Data Retrieval/Pushing Methods --------------------

    def _kp_uuid(self, path):
        raise NotImplementedError

    def _kp_exists(self, path, revision=None):
        raise NotImplementedError

    def _kp_status(self, path, revision=None, detailed=False):
        raise NotImplementedError

    def _kp_get_revision(self, path, status=None):
        raise NotImplementedError

    def _kp_get_revisions(self, path):
        raise NotImplementedError

    def _kp_read_ref(self, path, reference, revision=None):
        raise NotImplementedError

    def _kp_dir(self, path, parent=None, revision=None):
        raise NotImplementedError

    def _kp_has_ref(self, path, reference, revision=None):
        raise NotImplementedError

    def _kp_diff(self, path, head, base):
        raise NotImplementedError

    def _kp_write_ref(self, path, reference, data, uuid=None, revision=None):
        raise NotImplementedError

    def _kp_new_revision(self, path, uuid=None):
        raise NotImplementedError



import markdown
from flask import url_for
from knowledge_repo.post import KnowledgePost

MARKDOWN_EXTENSTIONS = ['markdown.extensions.extra',
                        'markdown.extensions.abbr',
                        'markdown.extensions.attr_list',
                        'markdown.extensions.def_list',
                        'markdown.extensions.fenced_code',
                        'markdown.extensions.footnotes',
                        'markdown.extensions.tables',
                        'markdown.extensions.smart_strong',
                        'markdown.extensions.admonition',
                        'markdown.extensions.codehilite',
                        'markdown.extensions.headerid',
                        'markdown.extensions.meta',
                        'markdown.extensions.sane_lists',
                        'markdown.extensions.smarty',
                        'markdown.extensions.toc(baselevel=3)',
                        'markdown.extensions.wikilinks']


def render_post_tldr(post):
    if isinstance(post, KnowledgePost):
        return markdown.Markdown(extensions=MARKDOWN_EXTENSTIONS).convert(post.headers.get('tldr'))
    else:
        return markdown.Markdown(extensions=MARKDOWN_EXTENSTIONS).convert(post.tldr)


def render_post_header(post):

    header_template = """
    <h1>{title}</h1>
    <p id='metadata'>
    <strong>Author(s): </strong>{author_hrefs} <br>
    <strong>Date Created</strong>: {date_created}<br>
    <strong>Date Updated</strong>: {date_updated}<br>
    <strong>Tags</strong><text>: </text><br>
    <strong>TLDR</strong>: {tldr}<br>
    </p>
    """

    if isinstance(post, KnowledgePost):
        return header_template.format(title=post.headers['title'],
                                      author_hrefs=', '.join(post.headers['authors']),
                                      date_created=post.headers['created_at'].isoformat(),
                                      date_updated=post.headers['updated_at'].isoformat(),
                                      tldr=render_post_tldr(post))
    else:
        author_hrefs = ', '.join(["<a href='{}'>{}</a>".format(
            url_for('index.render_feed', authors=author.username), author.format_name)
            for author in post.authors])
        return header_template.format(title=post.title,
                                      author_hrefs=author_hrefs,
                                      date_created=post.created_at.isoformat(),
                                      date_updated=post.updated_at.isoformat(),
                                      tldr=render_post_tldr(post))


def render_post_raw(post):
    if isinstance(post, KnowledgePost):
        return post.read().encode('ascii', 'ignore')
    else:
        return post.text.encode('ascii', 'ignore')


def render_post(post):
    """
    Renders the markdown as html
    """
    def intra_knowledge_urlmapper(name, url):
        if name == 'a' and url.startswith('knowledge:'):
            return url_for('render.render', markdown=url.split('knowledge:')[1])
        return None

    html = render_post_header(post) + (post if isinstance(post, KnowledgePost) else post.kp).to_string('html',
                                                                                                       skip_headers=True,
                                                                                                       urlmappers=[intra_knowledge_urlmapper])
    return html


def render_comment(comment):
    return markdown.Markdown().convert(comment.text)



import imp
import types
import logging
import os
import time

logger = logging.getLogger(__name__)


class KnowledgeRepositoryConfig(dict):

    def __init__(self, *args, **kwargs):
        super(KnowledgeRepositoryConfig, self).__init__(*args, **kwargs)
        self.DEFAULT_CONFIGURATION = {}

    def __getitem__(self, key):
        try:
            return super(KnowledgeRepositoryConfig, self).__getitem__(key)
        except:
            return self.DEFAULT_CONFIGURATION[key]

    def __getattr__(self, attr):
        return self[attr]

    def __setattr__(self, attr, value):
        self[attr] = value

    def __dir__(self):
        return list(set(list(self.DEFAULT_CONFIGURATION.keys()) + list(self.keys())))

    def update(self, *values, **kwargs):
        for value in values:
            if isinstance(value, dict):
                if 'DEFAULT_CONFIGURATION' in value:
                    value = value.copy()
                    value.pop('DEFAULT_CONFIGURATION')
                dict.update(self, value)
            elif isinstance(value, types.ModuleType):
                self.__update_from_module(value)
            elif type(value) == str:
                if os.path.exists(value):
                    self.__update_from_file(value)
                else:
                    logger.warning(
                        "Configuration file {} does not exist.".format(value))
            elif isinstance(value, type(None)):
                pass
            else:
                raise ValueError("Cannot interpret {}".format(value))
        dict.update(self, kwargs)

    def update_defaults(self, *values, **kwargs):
        for value in values:
            if type(value) == dict:
                self.DEFAULT_CONFIGURATION.update(value)
            elif isinstance(value, types.ModuleType):
                self.__defaults_from_module(value)
            elif type(value) == str:
                if os.path.exists(value):
                    self.__defaults_from_file(value)
                else:
                    logger.warning("Configuration file {} does not exist.".format(value))
            elif isinstance(value, type(None)):
                pass
            else:
                raise ValueError("Cannot interpret {}".format(value))
        self.DEFAULT_CONFIGURATION.update(kwargs)

    def __defaults_from_file(self, filename):
        self.__set_from_file(self.DEFAULT_CONFIGURATION, filename, force=True)

    def __update_from_file(self, filename):
        self.__set_from_file(self, filename)

    def __defaults_from_module(self, module):
        self.__set_from_module(self.DEFAULT_CONFIGURATION, module, force=True)

    def __update_from_module(self, module):
        self.__set_from_module(self, module)

    def __set_from_file(self, d, filename, force=False):
        config = imp.load_source('knowledge_repo.config_{}'.format(str(time.time()).replace('.', '')), filename)
        self.__set_from_module(d, config, force)

    def __set_from_module(self, d, module, force=False):
        for key in dir(module):
            if not key.startswith('_'):
                if not force and key not in self.DEFAULT_CONFIGURATION:
                    logger.debug(
                        "Ignoring configuration key `{}` which is not a valid configuration key.".format(key))
                else:
                    d[key] = getattr(module, key)



""" Define the routes that show all the posts.

This includes:
  - /feed
  - /cluster
  - /table
  - /favorites
"""
import os
import posixpath
from flask import request, render_template, redirect, Blueprint, current_app, make_response

from ..app import db_session
from ..utils.posts import get_posts
from ..models import Post, Tag, User, PageView
from ..utils.requests import from_request_get_feed_params
from ..utils.render import render_post_tldr

blueprint = Blueprint(
    'index', __name__, template_folder='../templates', static_folder='../static')


def has_no_empty_params(rule):
    defaults = rule.defaults if rule.defaults is not None else ()
    arguments = rule.arguments if rule.arguments is not None else ()
    return len(defaults) >= len(arguments)


@blueprint.route("/site-map")
@PageView.logged
def site_map():
    links = []
    for rule in current_app.url_map.iter_rules():
        # Filter out rules we can't navigate to in a browser
        # and rules that require parameters
        # if "GET" in rule.methods and has_no_empty_params(rule):
        # url = url_for(rule.endpoint, **(rule.defaults or {}))
        links.append((str(rule), rule.endpoint))
    # links is now a list of url, endpoint tuples
    return '<br />'.join(str(link) for link in links)


@blueprint.route('/')
@PageView.logged
def render_index():
    return redirect('/feed')


@blueprint.route('/favorites')
@PageView.logged
def render_favorites():
    """ Renders the index-feed view for posts that are liked """

    feed_params = from_request_get_feed_params(request)
    user_id = feed_params['user_id']

    user = (db_session.query(User)
            .filter(User.id == user_id)
            .first())
    posts = user.get_liked_posts

    post_stats = {post.path: {'all_views': post.view_count,
                              'distinct_views': post.view_user_count,
                              'total_likes': post.vote_count,
                              'total_comments': post.comment_count} for post in posts}

    return render_template("index-feed.html",
                           feed_params=feed_params,
                           posts=posts,
                           post_stats=post_stats,
                           top_header='Favorites',
                           contribs=current_app.config['plugins'])


@blueprint.route('/feed')
@PageView.logged
def render_feed():
    """ Renders the index-feed view """
    feed_params = from_request_get_feed_params(request)
    posts, post_stats = get_posts(feed_params)
    for post in posts:
        post.tldr = render_post_tldr(post)

    return render_template("index-feed.html",
                           feed_params=feed_params,
                           posts=posts,
                           post_stats=post_stats,
                           top_header='Knowledge Feed',
                           contribs=current_app.config['plugins'])


@blueprint.route('/table')
@PageView.logged
def render_table():
    """Renders the index-table view"""
    feed_params = from_request_get_feed_params(request)
    posts, post_stats = get_posts(feed_params)
    # TODO reference stats inside the template
    return render_template("index-table.html",
                           posts=posts,
                           post_stats=post_stats,
                           top_header="Knowledge Table",
                           feed_params=feed_params,
                           contribs=current_app.config['plugins'])


@blueprint.route('/cluster')
@PageView.logged
def render_cluster():
    """ Render the cluster view """
    # we don't use the from_request_get_feed_params because some of the
    # defaults are different
    filters = request.args.get('filters', '')
    sort_by = request.args.get('sort_by', 'alpha')
    group_by = request.args.get('group_by', 'folder')
    request_tag = request.args.get('tag')
    sort_desc = not bool(request.args.get('sort_asc', ''))

    post_query = db_session.query(Post).filter(Post.is_published)

    if filters:
        filter_set = filters.split(" ")
        for elem in filter_set:
            elem_regexp = "%," + elem + ",%"
            post_query = post_query.filter(Post.keywords.like(elem_regexp))

    if group_by == "author":
        author_to_posts = {}
        authors = (db_session.query(User).all())
        for author in authors:
            author_posts = [post for post in author.posts if post.is_published]
            if author_posts:
                author_to_posts[author.format_name] = author_posts
        tuples = [(k, v) for (k, v) in author_to_posts.items()]

    elif group_by == "tags":
        tags_to_posts = {}
        all_tags = (db_session.query(Tag).all())

        for tag in all_tags:
            tag_posts = [post for post in tag.posts if post.is_published]
            if tag_posts:
                tags_to_posts[tag.name] = tag.posts
        tuples = [(k, v) for (k, v) in tags_to_posts.items()]

    elif group_by == "folder":
        posts = post_query.all()
        # group by folder
        folder_to_posts = {}

        for post in posts:
            folder = posixpath.dirname(post.path)
            if folder in folder_to_posts:
                folder_to_posts[folder].append(post)
            else:
                folder_to_posts[folder] = [post]

        tuples = [(k, v) for (k, v) in folder_to_posts.items()]

    else:
        raise ValueError("Group by `{}` not understood.".format(group_by))

    if sort_by == 'alpha':
        grouped_data = sorted(tuples, key=lambda x: x[0])
    else:
        grouped_data = sorted(
            tuples, key=lambda x: len(x[1]), reverse=sort_desc)

    return render_template("index-cluster.html",
                           grouped_data=grouped_data,
                           filters=filters,
                           sort_by=sort_by,
                           group_by=group_by,
                           tag=request_tag,
                           contribs=current_app.config['plugins'])


@blueprint.route('/create')
@blueprint.route('/create/<knowledge_format>')
@PageView.logged
def create(knowledge_format=None):
    """ Renders the create knowledge view """
    if knowledge_format is None:
        return render_template("create-knowledge.html")

    cur_dir = os.path.dirname(os.path.realpath(__file__))
    knowledge_template = "knowledge_template.{}".format(knowledge_format)
    filename = os.path.join(cur_dir, '../../templates', knowledge_template)
    response = make_response(open(filename).read())
    response.headers["Content-Disposition"] = "attachment; filename=" + knowledge_template
    return response



import unittest
from bs4 import BeautifulSoup

from knowledge_repo import KnowledgeRepository


class TableTest(unittest.TestCase):

    def setUp(self):
        self.repo = KnowledgeRepository.for_uri('tests/test_repo', auto_create=True)
        self.app = self.repo.get_app(config='tests/config_server.py').test_client()
        self.headers = {}

    def test01_table_route(self):
        """
        test table view
        """
        rv = self.app.get("/table", headers=self.headers)
        assert (rv.status == "200 OK")
        soup = BeautifulSoup(rv.data.decode('utf-8'), 'html.parser')
        post_rows = soup.findAll('tr', {'class': 'table-post'})
        for row in post_rows:
            column_ids = ['path', 'author', 'tags', 'updated_at',
                          'views', 'upvotes', 'comments', 'tldr']

            for column in column_ids:
                elem = row.findAll(
                    'td', id=lambda x: x and x.startswith(column))
                assert elem
                if column == 'path':
                    path_link = elem[0].findAll("a")[0]
                    path_href = path_link['href']
                    assert self.app.get(
                        path_href, headers=self.headers).status == "200 OK"

    def test02_table_sorts(self):
        """
        Test sorting table view
        """
        sort_by = ['Title', 'CreatedAt', 'UpdatedAt',
                   'Uniqueviews', 'Upvotes', 'Comments']
        for sort in sort_by:
            rv = self.app.get('/table?sort_by=' + sort, headers=self.headers)
            assert (rv.status == "200 OK")
            rv = self.app.get('/table?sort_by=' + sort +
                              "&sort_asc=1", headers=self.headers)
            assert (rv.status == "200 OK")


if __name__ == '__main__':
    unittest.main()



""" Define the routes that add/delete comments on posts.

This includes:
  - /comment
  - /comments
  - /delete_comment
"""
import logging
from flask import request, Blueprint, g

from ..app import db_session
from ..models import Comment, Post, PageView
from ..utils.emails import send_comment_email

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

blueprint = Blueprint('comments', __name__,
                      template_folder='../templates', static_folder='../static')


@blueprint.route('/comment', methods=['POST'])
@PageView.logged
def post_comment():
    """ Post a comment underneath a post """

    path = request.args.get('path', '')
    comment_id = request.args.get('comment_id', '')
    data = request.get_json()

    post = (db_session.query(Post)
                      .filter(Post.path == path)
                      .first())

    if not post:
        raise Exception('Unable to find post')

    comment = (db_session.query(Comment)
                         .filter(Comment.id == comment_id)
                         .first())
    if not comment:
        comment = Comment(post_id=post.id)

    comment.text = data['text']
    comment.user_id = g.user.id
    db_session.add(comment)
    db_session.commit()

    send_comment_email(post_id=post.id,
                       commenter=g.user.format_name,
                       comment_text=data['text'])
    return ""


@post_comment.object_extractor
def post_comment():
    comment_id = request.args.get('comment_id', '')
    return {
        'id': comment_id if comment_id else None,
        'type': 'comment'
    }


@blueprint.route('/delete_comment')
@PageView.logged
def delete_comment():
    """ Delete a comment """
    try:
        comment_id = int(request.args.get('comment_id', ''))

        comments = (db_session.query(Comment)
                              .filter(Comment.id == comment_id)
                              .all())
        for comment in comments:
            # you can only delete your own comments - silently fail on others
            if comment.user_id == g.user.id:
                db_session.delete(comment)
        db_session.commit()
    except:
        logging.warning("ERROR processing request")
        pass

    return ""


@delete_comment.object_extractor
def delete_comment():
    comment_id = request.args.get('comment_id', '')
    return {
        'id': int(comment_id) if comment_id else None,
        'type': 'comment',
        'action': 'delete'
    }



from ..converter import KnowledgePostConverter


class MdConverter(KnowledgePostConverter):
    _registry_keys = ['md']

    def from_file(self, filename):
        with open(filename) as f:
            self.kp.write(f.read())



""" Define the routes that enable like/unlike on posts.

This includes:
  - /like
  - /unlike
"""
import logging
from flask import request, Blueprint, g
from sqlalchemy import and_

from ..app import db_session
from ..models import Post, Vote, PageView


blueprint = Blueprint(
    'vote', __name__, template_folder='../templates', static_folder='../static')


@blueprint.route('/like')
@PageView.logged
def like_post():
    """ Like a post """
    try:
        post_id = int(request.args.get('post_id', '-1'))  # This will prevent old code from adding invalid post_ids
        if post_id < 0:
            return "No Post Found to like!"
        vote = (db_session.query(Vote)
                .filter(and_(Vote.object_id == post_id, Vote.user_id == g.user.id))
                .first())
        if not vote:
            vote = Vote(user_id=g.user.id, object_id=post_id)
            db_session.add(vote)
            db_session.commit()
    except:
        logging.warning("ERROR processing request")
    return ""


@blueprint.route('/unlike')
@PageView.logged
def unlike_post():
    """ Un-like a post """
    try:
        post_id = int(request.args.get('post_id', '-1'))  # This will prevent old code from adding invalid post_ids
        if post_id < 0:
            return "No Post Found to Unlike!"
        votes = (db_session.query(Vote)
                 .filter(and_(Vote.object_id == post_id, Vote.user_id == g.user.id))
                 .all())
        if votes:
            for vote in votes:
                db_session.delete(vote)
            db_session.commit()
    except:
        logging.warning("ERROR processing request")
    return ""



"""Initial version

Revision ID: f8113482599e
Revises: None
Create Date: 2016-08-18 17:04:32.542965

"""

# revision identifiers, used by Alembic.
revision = '40dfdcb2b07c'
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('comments',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.Column('post_id', sa.Integer(), nullable=True),
    sa.Column('text', sa.Text(), nullable=True),
    sa.Column('type', sa.String(length=100), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('emails',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.Column('trigger_id', sa.Integer(), nullable=True),
    sa.Column('trigger_type', sa.String(length=100), nullable=True),
    sa.Column('object_id', sa.Integer(), nullable=True),
    sa.Column('object_type', sa.String(length=100), nullable=True),
    sa.Column('sent_at', sa.DateTime(), nullable=True),
    sa.Column('subject', sa.Text(), nullable=True),
    sa.Column('text', sa.Text(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('pageviews',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('page', sa.String(length=512), nullable=True),
    sa.Column('endpoint', sa.String(length=255), nullable=True),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.Column('object_id', sa.Integer(), nullable=True),
    sa.Column('object_type', sa.String(length=100), nullable=True),
    sa.Column('ip_address', sa.String(length=64), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('error_message', sa.Text(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('posts',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('path', sa.String(length=512), nullable=True),
    sa.Column('project', sa.String(length=512), nullable=True),
    sa.Column('repository', sa.String(length=512), nullable=True),
    sa.Column('revision', sa.Integer(), nullable=True),
    sa.Column('title', sa.Text(), nullable=True),
    sa.Column('tldr', sa.Text(), nullable=True),
    sa.Column('keywords', sa.Text(), nullable=True),
    sa.Column('thumbnail', sa.Text(), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('status', sa.Integer(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('path')
    )
    op.create_table('subscriptions',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.Column('object_id', sa.Integer(), nullable=True),
    sa.Column('object_type', sa.String(length=100), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('tags',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=500), nullable=True),
    sa.Column('description', sa.Text(), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('users',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('username', sa.String(length=500), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('votes',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.Column('object_id', sa.Integer(), nullable=True),
    sa.Column('object_type', sa.String(length=100), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('knowledge_post_author',
    sa.Column('post_id', sa.Integer(), nullable=True),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['post_id'], ['posts.id'],),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'],)
    )
    op.create_table('knowledge_post_tags',
    sa.Column('post_id', sa.Integer(), nullable=True),
    sa.Column('tag_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['post_id'], ['posts.id'],),
    sa.ForeignKeyConstraint(['tag_id'], ['tags.id'],)
    )


def downgrade():
    op.drop_table('knowledge_post_tags')
    op.drop_table('knowledge_post_author')
    op.drop_table('votes')
    op.drop_table('users')
    op.drop_table('tags')
    op.drop_table('subscriptions')
    op.drop_table('posts')
    op.drop_table('pageviews')
    op.drop_table('emails')
    op.drop_table('comments')



from builtins import object

import os
from functools import wraps

from .postprocessor import KnowledgePostProcessor
from .utils.registry import SubclassRegisteringABCMeta
from .utils.dependencies import check_dependencies
from future.utils import with_metaclass


def get_format(filename, format=None):
    if format is None:
        format = os.path.splitext(filename)[1]
        if format.startswith('.'):
            format = format[1:]
    return format


class KnowledgePostConverter(with_metaclass(SubclassRegisteringABCMeta, object)):
    _registry_keys = None  # File extensions

    def __init__(self, kp, format=None, postprocessors=None, **kwargs):
        check_dependencies(self.dependencies, "Whoops! You are missing some dependencies required to use `{}` instances.".format(self.__class__.__name__))
        self.kp = kp
        self.format = format
        if postprocessors is None:
            postprocessors = ['extract_images', 'format_checks']
        self.postprocessors = postprocessors
        self.init(**kwargs)

    @property
    def dependencies(self):
        return []

    def init(self):
        pass

    def __get_wrapped_with_postprocessors(self, f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            kp = self.kp
            f(*args, **kwargs)
            postprocessors = self.postprocessors

            if f.__name__ == 'from_file':
                filename = args[0] if len(args) > 0 else kwargs['filename']
                kp.orig_context = os.path.dirname(filename)

            if postprocessors is None:
                postprocessors = []
            for postprocessor in postprocessors:
                KnowledgePostProcessor._get_subclass_for(postprocessor).process(kp)

            return kp
        return wrapped

    def __getattribute__(self, attr):
        if attr in ['from_file', 'from_string']:
            return self.__get_wrapped_with_postprocessors(object.__getattribute__(self, attr))
        return object.__getattribute__(self, attr)

    def from_file(self, filename, **opts):
        raise NotImplementedError

    def from_string(self, string, **opts):
        raise NotImplementedError

    def to_file(self, filename, **opts):
        raise NotImplementedError

    def to_string(self, **opts):
        raise NotImplementedError

    @classmethod
    def for_file(cls, kp, filename, format=None, postprocessors=None):
        return cls.for_format(kp, get_format(filename, format), postprocessors=postprocessors)

    @classmethod
    def for_format(cls, kp, format, postprocessors=None):
        if format.lower() not in cls._registry:
            raise ValueError("The knowledge repository does not support files of type '{}'. Supported types are: {}."
                             .format(format, ','.join(list(cls._registry.keys()))))
        return cls._get_subclass_for(format.lower())(kp, format=format, postprocessors=postprocessors)



from builtins import object
from .utils.registry import SubclassRegisteringABCMeta
from future.utils import with_metaclass


class KnowledgePostProcessor(with_metaclass(SubclassRegisteringABCMeta, object)):
    _registry_keys = []

    @classmethod
    def process(cls, kp):
        return None



from abc import ABCMeta
import logging

logger = logging.getLogger(__name__)


class SubclassRegisteringABCMeta(ABCMeta):

    def __init__(cls, name, bases, dct):
        super(SubclassRegisteringABCMeta, cls).__init__(name, bases, dct)

        if not hasattr(cls, '_registry'):
            cls._registry = {}

        registry_keys = getattr(cls, '_registry_keys', [])
        if registry_keys:
            for key in registry_keys:
                if key in cls._registry and cls.__name__ != cls._registry[key].__name__:
                    logger.info("Ignoring attempt by class `{}` to register key '{}', which is already registered for class `{}`.".format(cls.__name__, key, cls._registry[key].__name__))
                else:
                    cls._registry[key] = cls

    def _get_subclass_for(cls, key):
        return cls._registry[key]



import posixpath

from ..repository import KnowledgeRepository


class MetaKnowledgeRepository(KnowledgeRepository):
    _registry_keys = None

    def init(self):
        pass

    def __repos_for_prefix(self, path_prefix=None):
        repo_prefixes = sorted(self.uri.keys(), key=lambda x: len(x), reverse=True)
        for repo_prefix in repo_prefixes:
            if path_prefix is None or path_prefix.startswith(repo_prefix):
                repo_prefix, repo, repo_path = self.__repo_for_path(path_prefix or repo_prefix)
                repo_path = None if not repo_path else repo_path
                yield repo_prefix, repo, repo_path

    def __repo_for_path(self, path):
        path_prefixes = sorted(self.uri.keys(), key=lambda x: len(x), reverse=True)
        for prefix in path_prefixes:
            if path.startswith(prefix):
                relpath = posixpath.relpath(path, prefix or '.') if path else ''
                return prefix, self.uri[prefix], relpath if relpath != '.' else None
        raise ValueError("No KnowledgeRepository found for '{}'.".format(path))

    def __distribute_method(self, method, **kwargs):
        return {name: getattr(repo, method)(**kwargs) for name, repo in list(self.uri.items())}

    def __distribute_attribute(self, attribute):
        return {name: getattr(repo, attribute) for name, repo in list(self.uri.items())}

    def __delegate_for_path(self, path, method, **kwargs):
        prefix, repo, path = self.__repo_for_path(path)
        return getattr(repo, method)(path=path, **kwargs)

    def __delegate_for_kp(self, kp, path, method, **kwargs):
        prefix, repo, path = self.__repo_for_path(path)
        return getattr(repo, method)(kp=kp, path=path, **kwargs)

    # ------------- Repository actions / state ------------------------------------

    def session_begin(self):
        return self.__distribute_method('session_begin')

    def session_end(self):
        return self.__distribute_method('session_end')

    @property
    def revision(self):
        return self.__distribute_attribute('revision')

    def update(self):
        return self.__distribute_method('update')

    def set_active_draft(self, path):
        return self.__delegate_for_path(path, 'set_active_draft')

    @property
    def status(self):
        return self.__distribute_attribute('status')

    @property
    def status_message(self):
        return self.__distribute_attribute('status_message')

    # -------------- Post retrieval methods --------------------------------------

    def _dir(self, prefix, statuses):
        # TODO: Handle masked paths
        for (repo_prefix, repo, nested_prefix) in self.__repos_for_prefix(prefix):
            for path in repo.dir(prefix=nested_prefix, status=statuses):
                yield posixpath.join(repo_prefix, path)

    # -------------- Post submission / addition user flow --------------------

    def _add_prepare(self, kp, path, update=False, **kwargs):
        return self.__delegate_for_kp(kp, path, '_add_prepare', update=update, **kwargs)

    def _add_cleanup(self, kp, path, update=False, **kwargs):
        return self.__delegate_for_kp(kp, path, '_add_cleanup', update=update, **kwargs)

    def _submit(self, path):  # Submit a post for review
        return self.__delegate_for_path(path, '_submit')

    def _accept(self, path):
        return self.__delegate_for_path(path, '_accept')

    def _publish(self, path):  # Publish a post for general perusal
        return self.__delegate_for_path(path, '_publish')

    def _unpublish(self, path):
        return self.__delegate_for_path(path, '_unpublish')

    def _remove(self, path, all=False):
        return self.__delegate_for_path(path, '_remove', all=all)

    # ----------- Knowledge Post Data Retrieval/Pushing Methods --------------------

    def _kp_repository_uri(self, path):
        prefix, repository, repo_path = self.__repo_for_path(path)
        return '{{{prefix}}}{repository_uri}'.format(prefix=prefix, repository_uri=repository._kp_repository_uri(repo_path))

    def _kp_uuid(self, path):
        return self.__delegate_for_path(path, '_kp_uuid')

    def _kp_exists(self, path, revision=None):
        return self.__delegate_for_path(path, '_kp_exists', revision=revision)

    def _kp_status(self, path, revision=None, detailed=False):
        return self.__delegate_for_path(path, '_kp_status', revision=revision, detailed=detailed)

    def _kp_is_published(self, path):
        return self.__delegate_for_path(path, '_kp_is_published')

    def _kp_get_revision(self, path):
        return self.__delegate_for_path(path, '_kp_get_revision')

    def _kp_get_revisions(self, path):
        return self.__delegate_for_path(path, '_kp_get_revisions')

    def _kp_read_ref(self, path, reference, revision=None):
        return self.__delegate_for_path(path, '_kp_read_ref', reference=reference, revision=revision)

    def _kp_dir(self, path, parent=None, revision=None):
        return self.__delegate_for_path(path, '_kp_dir', parent=parent, revision=revision)

    def _kp_has_ref(self, path, reference, revision=None):
        return self.__delegate_for_path(path, '_kp_has_ref', reference=reference, revision=revision)

    def _kp_diff(self, path, head, base):
        return self.__delegate_for_path(path, '_kp_diff', head=head, base=base)

    def _kp_write_ref(self, path, reference, data, uuid=None, revision=None):
        return self.__delegate_for_path(path, '_kp_write_ref', reference=reference, data=data, uuid=uuid, revision=revision)

    def _kp_new_revision(self, path, uuid=None):
        return self.__delegate_for_path(path, '_kp_new_revision', uuid=uuid)



import subprocess

__all__ = ['__author__', '__author_email__', '__version__', '__git_uri__', '__dependencies__', '__optional_dependencies__']

__author__ = "Nikki Ray (maintainer), Robert Chang, Dan Frank,  Chetan Sharma,  Matthew Wardrop"
__author_email__ = "nikki.ray@airbnb.com, robert.chang@airbnb.com, dan.frank@airbnb.com, chetan.sharma@airbnb.com, matthew.wardrop@airbnb.com"
__version__ = "0.6.3"
try:
    with open(os.devnull, 'w') as devnull:
        __version__ += '_' + subprocess.check_output(['git', 'rev-parse', 'HEAD'], shell=False, stderr=devnull).decode('utf-8').replace('\n', '')
except:
    pass
__git_uri__ = "git@github.com:airbnb/knowledge-repo.git"

# These are the core dependencies, and should include all packages needed for accessing repositories
# and running a non-server-side instance of the flask application. Optional dependencies for converters/etc
# should be defined elsewhere.
__dependencies__ = [
    # Knowledge Repository Dependencies
    'future',  # Python 2/3 support
    'enum34',  # Python 3.4+ enum object used for Post status
    'pyyaml',  # Yaml parser and utilities
    'markdown',  # Markdown conversion utilities
    'gitpython',  # Git abstraction
    'tabulate',  # Rendering information prettily in knowledge_repo script

    # Flask App Dependencies
    'flask',  # Main flask framework
    'flask_mail',  # Mail client and utilities
    'Flask-Migrate',  # Database migration utilities
    'sqlalchemy',  # Database abstractions
    'jinja2>=2.7',  # Templating engine
    'werkzeug',  # Development webserver
    'gunicorn',  # Deployed webserver
    'inflection',  # String transformation library
]

__optional_dependencies__ = {
    # ipynb notebook conversion suport
    'ipynb': [
        'nbformat',
        'nbconvert',
        'traitlets'
    ],

    # PDF to image conversion used by app
    'pdf': [
        'PyPDF2',  # image for parsing PDFs to images
        'wand',  # imagemagick integration for image uploading
    ],

    # Testing dependencies
    'dev': [
        'nose',  # Testing framework
        'beautifulsoup4',  # HTML/XML parser
        'coverage'  # Documentation coverage tester
    ]
}
__optional_dependencies__['all'] = [dep for deps in __optional_dependencies__.values() for dep in deps]



import imghdr
import io
import os

from knowledge_repo.utils.dependencies import check_dependencies
from knowledge_repo._version import __optional_dependencies__

ALLOWED_IMAGE_TYPES = ('png', 'jpeg', 'gif')


def get_file_extension(filepath):
    return os.path.splitext(filepath)[1]


def is_allowed_image_format(img_file):
    """ Checks if a given file is an image"""
    return imghdr.what(img_file) in ALLOWED_IMAGE_TYPES


def is_pdf(filename):
    return get_file_extension(filename) == '.pdf'


def pdf_page_to_png(src_pdf, pagenum=0, resolution=154):
    """
    Returns specified PDF page as wand.image.Image png.
    :param PyPDF2.PdfFileReader src_pdf: PDF from which to take pages.
    :param int pagenum: Page number to take.
    :param int resolution: Resolution for resulting png in DPI.
    """

    check_dependencies(__optional_dependencies__['pdf'])
    # Import libraries within this function so as to avoid import-time dependence
    import PyPDF2
    from wand.image import Image  # TODO: When we start using this again, document which system-level libraries are required.

    dst_pdf = PyPDF2.PdfFileWriter()
    dst_pdf.addPage(src_pdf.getPage(pagenum))

    pdf_bytes = io.BytesIO()
    dst_pdf.write(pdf_bytes)
    pdf_bytes.seek(0)

    img = Image(file=pdf_bytes, resolution=resolution)
    img.convert("png")

    return img



from __future__ import print_function
from builtins import input

import os
import shutil
import logging
import re
import git
import socket
from io import open

from knowledge_repo._version import __git_uri__
from ..repository import KnowledgeRepository
from ..utils.exec_code import get_module_for_source
from ..utils.types import str_types
from ..utils.encoding import encode

logger = logging.getLogger(__name__)


class GitKnowledgeRepository(KnowledgeRepository):
    _registry_keys = ['', 'git']

    @classmethod
    def create(cls, uri, embed_tooling=True):
        path = uri.replace('git://', '')
        if os.path.exists(path):
            response = input('Repository already exists. Do you want to convert it to a knowledge data repository? Note that this will override any existing `README.md` and `.knowledge_repo_config.py` files, and replace any submodule at `.resources`. (y/n) ')
            if response is not 'y':
                logger.warning('Not updating existing repository. Aborting!')
                return
        repo = git.Repo.init(path, mkdir=True)
        sm = None
        if embed_tooling is True or isinstance(embed_tooling, dict):
            if not isinstance(embed_tooling, dict):
                embed_tooling = {}
            tooling_repo = embed_tooling.get('repository', __git_uri__)
            tooling_branch = embed_tooling.get('branch', 'master')
            # Delete any existing submodule at .resources
            try:
                if '.resources' in repo.tree():
                    obj = repo.tree()['.resources']
                    if isinstance(obj, git.Submodule):
                        sm = repo.tree()['.resources']
                        sm._name = 'embedded_knowledge_repo'
                        sm.remove()
                    else:
                        repo.git.rm(obj.path)
            except ValueError:  # Repository has no active refs
                pass
            sm = repo.create_submodule(name='embedded_knowledge_repo', path='.resources', url=tooling_repo, branch=tooling_branch)
        shutil.copy(os.path.join(os.path.dirname(__file__), '../config_defaults.py'),
                    os.path.join(path, '.knowledge_repo_config.py'))
        shutil.copy(os.path.join(os.path.dirname(__file__), '../templates', 'repo_data_readme.md'),
                    os.path.join(path, 'README.md'))
        repo.index.add(['.knowledge_repo_config.py', 'README.md'])
        repo.index.commit("Initial creation of knowledge data repository structure.")
        if sm is not None:
            repo.submodule('embedded_knowledge_repo').update()
        return GitKnowledgeRepository(path)

    def init(self, config='git:////.knowledge_repo_config.py', auto_create=False):
        self.config.update_defaults(published_branch='master')
        self.config.update_defaults(remote_name='origin')
        self.auto_create = auto_create
        self.path = self.uri.replace('git://', '')

        if config.startswith('git:////'):
            self.config.update(get_module_for_source(self.git_read(config.replace('git:////', '')), 'git_config'))
        else:
            self.config.update(os.path.join(self.path, config))

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, path):
        assert isinstance(path, str), "The path specified must be a string."
        path = os.path.abspath(os.path.expanduser(path))
        if not os.path.exists(path):
            path = os.path.abspath(path)
            if self.auto_create:
                self.create(path)
            else:
                raise ValueError("Provided path '{}' does not exist.".format(path))
        assert self.__is_valid_repo(
            path), "Provided path '{}' is not a valid repository.".format(path)
        self._path = path
        self.uri = path  # Update uri to point to absolute path of repository

    def __is_valid_repo(self, path):
        try:
            git.Repo(path)
            return True
        except git.InvalidGitRepositoryError:
            return False

    @property
    def git(self):
        if not hasattr(self, '_git'):
            self._git = git.Repo(self.path)
        return self._git

    @property
    def git_has_remote(self):
        return hasattr(self.git.remotes, self.config.remote_name)

    @property
    def git_remote(self):
        if self.git_has_remote:
            return self.git.remote(self.config.remote_name)
        return None

    # ----------- Repository actions / state ------------------------------------
    @property
    def revision(self):
        return self.git.commit().hexsha

    def update(self, branch=None):
        branch = branch or self.config.published_branch
        if not self.git_has_remote:
            return
        if not self.__remote_available:
            logger.warning("Cannot connect to remote repository hosted on {}. Continuing locally with potentially outdated code.".format(
                self.__remote_host))
            return
        logger.info("Fetching updates to the knowledge repository...")
        self.git_remote.fetch()
        current_branch = self.git.active_branch
        self.git.branches.master.checkout()
        self.git_remote.pull(branch)
        try:
            sm = self.git.submodule('embedded_knowledge_repo')
        except ValueError:  # This repository does not use embedded knowledge_repo tools or it is misnamed
            # Check for misnamed submodule
            sm = None
            for submodule in self.git.submodules:
                if submodule.path == '.resources':
                    sm = submodule
                    break
        if sm is not None:
            sm_target_url = sm.config_reader().get_value('url')
            sm_actual_url = sm.module().git.config('--get', 'remote.origin.url')
            if sm_target_url != sm_actual_url:
                logging.info('Migrating embedded tooling from {} to {}.'.format(sm_actual_url, sm_target_url))
                self.git.git.submodule('sync')
                self.git.git.submodule('update', '--init', '--recursive', '--remote', '--force', '--checkout')
            sm.update(init=True, force=True)
        current_branch.checkout()

    def set_active_draft(self, path):  # TODO: deprecate
        branch = self.git_branch_for_post(self._kp_path(path))
        self.config.published_branch = branch.name
        branch.checkout()

    @property
    def status(self):
        return {
            'branch': self.git.active_branch.name,
            'changed_files': [str(diff.a_path) for diff in self.git_diff()]
        }

    @property
    def status_message(self):
        status = self.status
        message = "Currently checked out on the `{branch}` branch.".format(branch=status['branch'])
        if len(status['changed_files']) > 0:
            message += "Files modified: \n {modified}".format(modified='\n\t- '.join(status['changed_files']))
        return message

    # ---------------- Git properties and actions -------------------------
    def git_dir(self, prefix=None, commit=None):
        commit = self.git.commit(commit or self.config.published_branch)
        tree = commit.tree
        if prefix is not None:
            tree = tree[prefix]
        return [o.path for o in tree.traverse(
                prune=lambda i, d: isinstance(i, git.Submodule) or os.path.dirname(i.path).endswith('.kp'),
                visit_once=False,
                predicate=lambda i, d: i.path.endswith('.kp')
                )
                ]

    def git_read(self, path, commit=None):
        commit = self.git.commit(commit or self.config.published_branch)
        return commit.tree[path].data_stream.read()

    @property
    def git_local_branches(self):
        unmerged_branches = [branch.replace(
            '*', '').strip() for branch in self.git.git.branch('--no-merged', self.config.published_branch).split('\n')]
        return unmerged_branches

    def __get_path_from_ref(self, ref):
        refs = ref.split('/')
        for i, ref in enumerate(refs):
            if ref.endswith('.kp'):
                break
        if not ref.endswith('.kp'):
            return None
        return '/'.join(refs[:i + 1])

    def git_local_posts(self, branches=None, as_dict=False):
        if branches is None:
            branches = self.git_local_branches
        posts = {}
        for branch in branches:
            posts[branch] = set([self.__get_path_from_ref(diff.a_path) for diff in self.git_diff(branch)])
            posts[branch].discard(None)

        if not as_dict:
            out_posts = set()
            for branch, ps in posts.items():
                out_posts.update(ps)
            return list(out_posts)
        return posts

    def git_branch_for_post(self, path, interactive=False):
        if path is None:
            return None
        if path in self.git_local_branches:
            return self.git_branch(path)

        branches = []
        for branch in self.git_local_branches:
            if path in self.git_local_posts(branches=[branch]):
                branches.append(branch)

        if len(branches) == 0:
            if path in self.dir():
                return self.git_branch(self.config.published_branch)
            return None

        if len(branches) == 1:
            return self.git_branch(branches[0])

        # Deal with ambiguity
        if interactive:
            print("There are multiple branches for post '{}'.".format(path))
            for i, branch in enumerate(branches):
                print("{}. {}".format(i, branch))
            response = None
            while not isinstance(response, int):
                response = input('Please select the branch you would like to use: ')
                try:
                    response = int(response)
                except:
                    response = None
        else:
            response = 0

        return self.git_branch(branches[response])

    def git_branch(self, branch=None):
        if isinstance(branch, git.Head):
            return branch
        if branch is None:
            return self.git.active_branch

        if not isinstance(branch, str_types):
            raise ValueError("'{}' of type `{}` is not a valid branch descriptor.".format(branch, type(branch)))

        try:
            return self.git.branches[branch]
        except IndexError:
            raise ValueError("Specified branch `{}` does not exist.".format(branch))

    def git_checkout(self, branch, soft=False, reset=False, create=False):
        if not create:
            branch_obj = self.git_branch(branch)
            branch_obj.checkout()
            return branch_obj

        if soft and self.git.active_branch.name not in [self.config.published_branch, branch] and not self.git.active_branch.name.endswith('.kp'):
            response = None
            while response not in ['y', 'n']:
                response = input('It looks like you have checked out the `{}` branch, whereas we were expecting to use `{}`. Do you want to use your current branch instead? (y/n) '.format(self.git.active_branch.name, branch))
                if response == 'y':
                    branch = self.git.active_branch.name

        if reset or branch not in [b.name for b in self.git.branches]:
            ref_head = None
            if self.git_has_remote:
                for ref in self.git_remote.refs:
                    if ref.name == branch:
                        ref_head = ref
                        break
            if not ref_head:
                ref_head = self.git_remote.refs.master if self.git_has_remote else self.git.branches.master
            else:
                logger.warning(
                    "The branch `{}` already exists as upstream, and you maybe clobbering someone's work. Please check.".format(ref_head.name))
            branch = self.git.create_head(branch, ref_head, force=True)
        else:
            branch = self.git_branch(branch)
        branch.checkout()
        return branch

    def git_diff(self, ref=None, ref_base=None, patch=False):
        commit = self.git.commit(ref)
        ref = self.git.merge_base(self.git.commit(ref_base or self.config.published_branch), commit)[0]
        return commit.diff(ref, create_patch=patch)

    # ---------------- Post retrieval methods --------------------------------

    def _dir(self, prefix, statuses):
        posts = set()

        if any([status != self.PostStatus.PUBLISHED for status in statuses]):
            local_posts = self.git_local_posts(as_dict=True)

        for status in statuses:
            if status == self.PostStatus.PUBLISHED:
                posts.update(self.git_dir(prefix=prefix, commit=self.config.published_branch))
            else:
                for branch in local_posts:
                    for post_path in local_posts[branch]:
                        if prefix is not None and not post_path.startswith(prefix):
                            continue
                        if self._kp_status(post_path, branch=branch) in statuses:
                            posts.add(post_path)
        for post in sorted(posts):
            yield post

    # ------------- Post submission / addition user flow ----------------------
    def _add_prepare(self, kp, path, update=False, branch=None, squash=False, message=None):
        target = os.path.abspath(os.path.join(self.path, path))
        if self.git_has_remote:
            branch = branch or path
        else:
            logger.warning("This repository does not have a remote, and so post review is being skipped. Adding post directly into published branch...")
            branch = self.config.published_branch

        # Create or checkout the appropriate branch for this project
        logger.info("Checking out (and/or creating) a new branch `{}`...".format(branch))
        branch_obj = self.git_checkout(branch, soft=True, reset=squash, create=True)
        branch = branch_obj.name

        # Verify that post path does not exist (unless we are updating the post)
        assert update or not os.path.exists(target), "A knowledge post already exists at '{}'! If you wanted to update it, please pass the '--update' flag.".format(path)

        # Add knowledge post to local branch
        logger.info("Adding and committing '{}' to local branch `{}`...".format(path, branch))

    def _add_cleanup(self, kp, path, update=False, branch=None, squash=False, message=None):
        self.git.index.add([path])

        # Commit the knowledge post and rollback if it fails
        try:
            if message is None:
                message = input("Please enter a commit message for this post: ")
            self.git.index.commit(message)
        except (KeyboardInterrupt, Exception) as e:
            if message is None:
                logger.warning("No commit message input for post '{}'. Rolling back post addition...")
            else:
                logger.error("Something went wrong. Rolling back post addition...")
            self.git.index.reset()
            try:
                self.git.git.clean('-df', path)
                self.git.git.checkout('--', path)
            except:
                pass
            raise e

    def _submit(self, path=None, branch=None, force=False):
        if not self.git_has_remote:
            raise RuntimeError("Could not find remote repository `{}` into which this branch should be submitted.".format(self.config.remote_name))
        if branch is None and path is None:
            raise ValueError("To submit a knowledge post, a path to the post and/or a git branch must be specified.")
        if branch is None:
            branch = self.git_branch_for_post(path)
        if branch is None:
            raise ValueError("It does not appear that you have any drafts in progress for '{}'.".format(path))

        if not self.__remote_available:
            raise RuntimeError("Cannot connect to remote repository {} ({}). Please check your connection, and then try again.".format(self.config.remote_name, self.git_remote.url))

        self.git_remote.push(branch, force=force)
        logger.info("Pushed local branch `{}` to upstream branch `{}`. Please consider starting a pull request, or otherwise merging into master.".format(branch, branch))

    def _publish(self, path):  # Publish a post for general perusal
        raise NotImplementedError

    def _unpublish(self, path):  # unpublish a post for general perusal
        raise NotImplementedError

    def _accept(self, path):  # Approve to publish a post for general perusal
        pass

    def _remove(self, path, all=False):
        raise NotImplementedError

    # ------------ Knowledge Post Data Retrieval Methods -------------------------

    def _kp_uuid(self, path):
        try:
            return self._kp_read_ref(path, 'UUID')
        except:
            return None

    def _kp_path(self, path, rel=None):
        return KnowledgeRepository._kp_path(self, os.path.expanduser(path), rel=rel or self.path)

    def _kp_exists(self, path, revision=None):
        # For speed, first check whether it exists in the checked out branch, then search more deeply
        return os.path.isdir(os.path.join(self.path, path)) or (self.git_branch_for_post(path, interactive=False) is not None)

    def _kp_status(self, path, revision=None, detailed=False, branch=None):
        if not hasattr(self, '_dir_cache'):
            self._dir_cache = {path: None for path in self.dir()}
        if path in self._dir_cache:
            return self.PostStatus.PUBLISHED
        if branch is None:
            branch = self.git_branch_for_post(path, interactive=False)
        else:
            branch = self.git_branch(branch)

        if branch is None:
            return ValueError("No such post: {}".format(path))

        if branch.name == self.config.published_branch:
            status = self.PostStatus.PUBLISHED, None
        elif self.git_has_remote and branch.name in self.git_remote.refs:
            remote_branch = self.git_remote.refs[branch.name].name
            behind = len(list(self.git.iter_commits('{}..{}'.format(branch, remote_branch))))
            ahead = len(list(self.git.iter_commits('{}..{}'.format(remote_branch, branch))))

            status = (self.PostStatus.SUBMITTED,
                      (" - {} commits behind".format(behind) if behind else '') +
                      (" - {} commits ahead".format(ahead) if ahead else '') +
                      (" [On branch: {}]".format(branch) if branch != path else ''))
        else:
            status = self.PostStatus.DRAFT, None

        if detailed:
            return status
        return status[0]

    def _kp_get_revision(self, path):
        # We use a 'REVISION' file in the knowledge post folder rather than using git
        # revisions because using git rev-parse is slow.
        try:
            return int(self._kp_read_ref(path, 'REVISION'))
        except:
            return 0

    def _kp_get_revisions(self, path):  # slow
        # TODO: In the future, we may want to use something like:
        #    self.git.iter_commits(paths=os.path.join(self.path, path, 'knowledge.md'))
        # But this will require a lot of piping and may not make sense in the context
        # of a non-bare git repository.
        raise NotImplementedError

    def _kp_write_ref(self, path, reference, data, uuid=None, revision=None):
        ref_path = os.path.join(self.path, path, reference)
        ref_dir = os.path.dirname(ref_path)
        if not os.path.exists(ref_dir):
            os.makedirs(ref_dir)
        with open(ref_path, 'wb') as f:
            return f.write(data)

    def _kp_dir(self, path, parent=None, revision=None):  # TODO: Account for revision
        if parent:
            path = os.path.join(path, parent)
        for dirpath, dirnames, filenames in os.walk(os.path.join(self.path, path)):
            for filename in filenames:
                if dirpath == "" and filename == "REVISION":
                    continue
                yield os.path.relpath(os.path.join(dirpath, filename), os.path.join(self.path, path))

    def _kp_has_ref(self, path, reference, revision=None):  # TODO: Account for revision
        return os.path.isfile(os.path.join(self.path, path, reference))

    def _kp_diff(self, path, head, base):
        raise NotImplementedError

    def _kp_new_revision(self, path, uuid=None):
        self._kp_write_ref(path, "REVISION", encode(self._kp_get_revision(path) + 1))
        if uuid:
            self._kp_write_ref(path, "UUID", encode(uuid))

    def _kp_read_ref(self, path, reference, revision=None):
        with open(os.path.join(self.path, path, reference), 'rb') as f:
            return f.read()

    # ------------- Utility methods --------------------------------------
    def __abspath(self, path):
        return os.path.abspath(os.path.join(self.path, path))

    @property
    def __remote_host(self):
        if self.git_has_remote:
            # TODO: support more types of hosts
            m = re.match('.*?@(.*?):\.*?', self.git_remote.url)
            if m:  # shorthand ssh uri
                return m.group(1)
        return None

    @property
    def __remote_available(self):
        # TODO: support more types of hosts
        host = self.__remote_host

        if host:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(0.5)
            try:
                s.connect((socket.gethostbyname(host), 22))
                return True
            except:
                return False
            finally:
                s.close()
        return True



import os
import re
import logging

from ..postprocessor import KnowledgePostProcessor

logger = logging.getLogger(__name__)


class ExtractImages(KnowledgePostProcessor):
    _registry_keys = ['extract_images']

    @classmethod
    def process(cls, kp):
        images = cls.find_images(kp.read())
        cls.collect_images(kp, images)
        cls.cleanup(kp)

    @classmethod
    def find_images(cls, md):
        images = []
        images.extend(cls.collect_images_for_pattern(
            md, '<img.*src=[\'"](.*?)[\'"].*>'))
        images.extend(cls.collect_images_for_pattern(md, '\!\[.*\]\((.*)\)'))
        return sorted(images, key=lambda x: x['offset'])

    @classmethod
    def collect_images_for_pattern(cls, md, pattern=None):
        p = re.compile(pattern)
        return [{'offset': m.start(), 'tag': m.group(0), 'src': m.group(1)} for m in p.finditer(md)]

    @classmethod
    def collect_images(cls, kp, images):
        if len(images) == 0:
            return
        md = kp.read()
        images = images[::-1]
        for image in images:
            if cls.skip_image(kp, image):
                continue
            orig_path = os.path.join(kp.orig_context, image['src'])

            new_path = None
            if kp._has_ref(image['src']):
                new_path = cls.copy_image(kp, image['src'], is_ref=True)
            elif os.path.exists(orig_path):
                new_path = cls.copy_image(kp, orig_path)
            else:
                logger.warning("Could not find an image at: {}".format(image['src']))
            if not new_path:
                continue
            md = cls.replace_image_locations(md, image['offset'], image['tag'], image['src'], new_path)
        kp.write(md)

    @classmethod
    def skip_image(cls, kp, image):
        if re.match('http[s]?://', image['src']):
            return True
        if image['src'].startswith('images/') and image['src'] in kp.image_paths:
            return True
        return False

    @classmethod
    def copy_image(cls, kp, path, is_ref=False):
        if is_ref:
            return
        with open(path) as f:
            kp.write_image(os.path.basename(path), f.read())
        return os.path.join('images', os.path.basename(path))

    @classmethod
    def replace_image_locations(cls, md, offset, match, old_path, new_path):
        pre = md[:offset]
        post = md[offset + len(match):]
        return pre + match.replace(old_path, new_path) + post

    @classmethod
    def cleanup(cls, kp):
        pass



"""Test tag related functionality."""
import datetime
import json
import unittest

from knowledge_repo import KnowledgeRepository
from knowledge_repo.app.proxies import db_session
from knowledge_repo.app.models import Tag, User, Subscription, Post, assoc_post_tag


class TagsTest(unittest.TestCase):
    """The Test Class for all tag functionality."""

    def setUp(self):
        """Set up the TagsTest Class by instantiating a tag."""
        self.repo = KnowledgeRepository.for_uri('tests/test_repo', auto_create=True)
        self.app = self.repo.get_app(config='tests/config_server.py')
        self.client = self.app.test_client()

        # we'll create a tag for the purpose of tag testing
        test_run = datetime.datetime.now().strftime('%Y-%m-%d')

        with self.app.app_context():
            tag = Tag(name='test/tag_' + test_run)
            db_session.commit()
            TagsTest.tag_id = tag.id
            TagsTest.tag_name = tag.name

            # the default user is "knowledge_default"
            user = (db_session.query(User)
                    .filter(User.username == 'knowledge_default')
                    .first())

            TagsTest.user_id = user.id

            # We need an example post as well to test some of the
            # add/delete/rename functionality
            example_post = (db_session.query(Post).first())
            TagsTest.post_path = example_post.path

    def test01_test_subscribe_to_tag(self):
        """Test the subscription feature."""
        headers = {"X-InternalAuth-Username": "knowledge_default"}

        route_name = '/toggle_tag_subscription'
        tag_name = 'tag_name=' + TagsTest.tag_name
        action = 'subscribe_action=subscribe'
        full_url = '{route_name}?{tag_name}&{action}'.format(**locals())
        # check that status == 200 & the tag has been subscribed too
        rv = self.client.post(full_url, headers=headers)
        assert (rv.status == "200 OK")

        with self.app.app_context():
            subscription = (db_session.query(Subscription)
                            .filter(Subscription.object_id == TagsTest.tag_id)
                            .filter(Subscription.object_type == 'tag')
                            .first())
            assert subscription

    def test02_test_unsubscribe_to_tag(self):
        """Test the unsubscribe feature."""
        headers = {"X-InternalAuth-Username": "knowledge_default"}

        route_name = '/toggle_tag_subscription'

        with self.app.app_context():
            tag = db_session.query(Tag).first()

            tag_name = 'tag_name=' + TagsTest.tag_name
            action = 'subscribe_action=unsubscribe'
            full_url = '{route_name}?{tag_name}&{action}'.format(
                **locals())

            # check that status == 200 & the tag has been subscribed too
            rv = self.client.post(full_url, headers=headers)
            assert (rv.status == "200 OK")

            subscription = (db_session.query(Subscription)
                            .filter(Subscription.object_id == TagsTest.tag_id)
                            .filter(Subscription.object_type == 'tag')
                            .first())
            assert not subscription

    @unittest.skip("Current model code is broken: format_tags does not work, and does not make sense.")
    def test03_test_tag_list(self):
        """/tag_list is called when someone changes tags on the UI.

        Add the test tag to a post and ensure is has been added
        to the post_obj
        """
        with self.app.app_context():
            # we need to get the tag_string associate with the original post
            post_obj = (db_session.query(Post)
                        .filter(Post.path == TagsTest.post_path)
                        .first())
            post_tags_string = post_obj.format_tags()

            # add test/tag to this string
            test_tag_string = TagsTest.tag_name
            post_tags_string += ", " + test_tag_string

            data = {"tags": post_tags_string}
            rv = self.client.post('/tag_list?post_id=' + TagsTest.post_path,
                                  data=json.dumps(data),
                                  content_type='application/json',
                                  method='POST')

            assert rv.status == "200 OK"

            # check the new post tags string, and assert that "test/tag"
            post_obj = (db_session.query(Post)
                        .filter(Post.path == TagsTest.post_path)
                        .first())
            post_tags_string = post_obj.format_tags()

            assert test_tag_string in post_tags_string

    def test04_rename_tag_post(self):
        """/rename_tag is called from the batch_edit tag page.

        Rename the test/tag tag to test/tag_renamed, and
        check to make sure that old tag is deleted (we'll re-add it),
        the new tag is added, and
        the tag_string for related posts has changed
        """
        new_tag_string = TagsTest.tag_name + '_renamed'

        with self.app.app_context():
            # get all the posts with the old_tag_string
            post_tags = (db_session.query(assoc_post_tag)
                         .filter(assoc_post_tag.c.tag_id == TagsTest.tag_id)
                         .all())
            post_ids = [x[0] for x in post_tags]

            data = {'oldTagId': TagsTest.tag_id,
                    'newTag': new_tag_string}

            rv = self.client.post('/rename_tag',
                                  data=json.dumps(data),
                                  content_type='application/json')

            assert(rv.status == "200 OK")

            old_tag_obj = (db_session.query(Tag)
                           .filter(Tag.id == TagsTest.tag_id)
                           .all())
            assert not old_tag_obj

            new_tag_obj = (db_session.query(Tag)
                           .filter(Tag.name == new_tag_string)
                           .all())

            assert new_tag_obj and len(new_tag_obj) == 1
            new_id = new_tag_obj[0].id

            for post_id in post_ids:
                post_tag = (db_session.query(assoc_post_tag)
                            .filter(assoc_post_tag.c.post_id == post_id)
                            .filter(assoc_post_tag.c.tag_id == new_id)
                            .all())
                assert post_tag

    def test05_delete_tag(self):
        """/delete_tag_post deletes a tag from all posts.

        Check to make sure the tag is deleted, and all the posts with
        that tag don't have an associate
        """
        with self.app.app_context():
            tag_id = Tag(name=TagsTest.tag_name + '_renamed').id
            post_tags = (db_session.query(assoc_post_tag)
                         .filter(assoc_post_tag.c.tag_id == tag_id)
                         .all())
            post_ids = [x[0] for x in post_tags]
            delete_tag_url = '/delete_tag_post?tag_id={tag_id}'.format(**locals())
            rv = self.client.get(delete_tag_url)

            assert rv.status == "200 OK"

            tag = (db_session.query(Tag)
                   .filter(Tag.name == TagsTest.tag_name + '_renamed')
                   .first())

            assert not tag

            for post_id in post_ids:
                post_tag = (db_session.query(assoc_post_tag)
                            .filter(assoc_post_tag.c.tag_id == tag_id)
                            .filter(assoc_post_tag.c.post_id == post_id)
                            .all())
                assert not post_tag

    def tearDown(self):
        """Delete the tag that we created."""
        with self.app.app_context():
            tag = (db_session.query(Tag)
                   .filter(Tag.name == TagsTest.tag_name)
                   .first())
            if tag:
                db_session.delete(tag)
                db_session.commit()

if __name__ == '__main__':
    unittest.main()
